from base64 import b64decode as license64

from BroTools.core.broqt.vendor.Qt import QtCore, QtGui, QtWidgets
from BroTools.core.broqt.color import ColorPalette
from typing import List, Tuple, Optional
import math
from BroTools.core.utils import lerp
from BroTools.core.curves import CurvePoint
from functools import partial

class CurvePreviewWidget(QtWidgets.QWidget):
    'Small widget that shows a preview of the curve'
    clicked = QtCore.Signal()

    def __init__(A, parent=None):
        super().__init__(parent)
        A.points = []
        A.setMinimumSize(120, 30)
        A.setMaximumSize(9999, 60)

    def set_points(C, points):
        B = []
        for A in points:
            if isinstance(A, CurvePoint):
                B.append(A)
            elif (isinstance(A, (list, tuple)) and (len(A) == 3)):
                B.append(CurvePoint(*A))
            else:
                raise ValueError('Invalid point format')
        C.points = sorted(B, key=(lambda x: x.position))
        C.update()

    def mousePressEvent(A, event):
        if (event.button() == QtCore.Qt.LeftButton):
            A.clicked.emit()

    def paintEvent(A, event):
        if (not A.points):
            return
        G = QtGui.QPainter(A)
        G.setRenderHint(QtGui.QPainter.Antialiasing)
        G.fillRect(A.rect(), QtGui.QColor(30, 30, 30))
        H = QtGui.QPen(QtGui.QColor(201, 145, 38))
        H.setWidth(1)
        H.setStyle(QtCore.Qt.DashLine)
        G.setPen(H)
        (K, B) = (A.width(), A.height())
        F = QtGui.QPainterPath()
        I = min((A.value for A in A.points))
        X = max((A.value for A in A.points))
        D = (X - I)
        P = A.points[0]
        Y = (P.position * K)
        Z = (B - ((((P.value - I) / D) * B) if D else (B / 2)))
        F.moveTo(Y, Z)
        for J in range((len(A.points) - 1)):
            Q = A.points[J]
            L = A.points[(J + 1)]
            a = (Q.position * K)
            b = (B - ((((Q.value - I) / D) * B) if D else (B / 2)))
            c = (L.position * K)
            d = (B - ((((L.value - I) / D) * B) if D else (B / 2)))
            E = QtCore.QPointF(a, b)
            C = QtCore.QPointF(c, d)
            if (L.interpolation == CurvePoint.LINEAR):
                F.lineTo(C)
            elif (L.interpolation == CurvePoint.SMOOTH):
                N = (E + QtCore.QPointF(((C.x() - E.x()) * 0.5), 0))
                O = (C - QtCore.QPointF(((C.x() - E.x()) * 0.5), 0))
                F.cubicTo(N, O, C)
            elif (L.interpolation == CurvePoint.SPLINE):
                M = 0.5
                R = E
                if (J > 0):
                    S = A.points[(J - 1)]
                    e = (S.position * K)
                    f = (B - ((((S.value - I) / D) * B) if D else (B / 2)))
                    R = QtCore.QPointF(e, f)
                T = C
                if (J < (len(A.points) - 2)):
                    U = A.points[(J + 2)]
                    g = (U.position * K)
                    h = (B - ((((U.value - I) / D) * B) if D else (B / 2)))
                    T = QtCore.QPointF(g, h)
                V = (C.x() - E.x())
                i = (E.x() + (V * M))
                j = (E.y() + (((C.y() - R.y()) * M) * 0.5))
                k = (C.x() - (V * M))
                l = (C.y() - (((T.y() - E.y()) * M) * 0.5))
                N = QtCore.QPointF(i, j)
                O = QtCore.QPointF(k, l)
                F.cubicTo(N, O, C)
            else:
                F.lineTo(QtCore.QPointF(C.x(), E.y()))
                F.lineTo(C)
        G.drawPath(F)
        H = QtGui.QPen(QtGui.QColor(75, 74, 74))
        H.setWidth(1)
        G.setPen(H)
        m = A.rect().adjusted(0, 0, (- 1), (- 1))
        W = QtGui.QPainterPath()
        W.addRoundedRect(m, 2, 2)
        G.drawPath(W)

class CurvePreviewWidgetWithEditor(CurvePreviewWidget):
    valueChanged = QtCore.Signal(list)

    def __init__(A, points=[CurvePoint(1, 1, CurvePoint.LINEAR), CurvePoint(0, 0, CurvePoint.LINEAR)], parent=None):
        B = points
        super().__init__(parent)
        A.editor = CurveEditorWindow(B)
        A.editor.set_points(B)
        A.editor.curve_changed.connect(A._on_curve_changed)
        A.set_points(B)
        A.clicked.connect(A._open_editor)
        A.updating = False

    def _open_editor(A):
        A.editor.show()

    def _on_curve_changed(A, points):
        B = points
        if A.updating:
            return
        A.set_points(B)
        A.valueChanged.emit(B)

    def value(A):
        return A.points

    def setValue(A, value):
        B = value
        A.updating = True
        A.set_points(B)
        A.valueChanged.emit(B)
        A.editor.set_points(B)
        A.updating = False

class ToggleableCurveEditor(QtWidgets.QWidget):
    'A wrapper that toggles between CurvePreviewWidget and CurveEditorWindow in place'
    valueChanged = QtCore.Signal(list)

    def __init__(A, points=None, parent=None, min_value=None, max_value=None):
        B = points
        super().__init__(parent)
        if (B is None):
            B = [CurvePoint(0, 0, CurvePoint.LINEAR), CurvePoint(1, 1, CurvePoint.LINEAR)]
        A.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        A.layout = QtWidgets.QStackedLayout(A)
        A.layout.setContentsMargins(0, 0, 0, 0)
        A.preview = CurvePreviewWidget(A)
        A.preview.set_points(B)
        A.preview.clicked.connect(A._show_editor)
        A.editor_container = QtWidgets.QWidget(A)
        C = QtWidgets.QVBoxLayout(A.editor_container)
        C.setContentsMargins(0, 0, 0, 0)
        A.editor = CurveEditorWindow(B, min_value, max_value, A)
        C.addWidget(A.editor)
        D = QtWidgets.QHBoxLayout()
        D.addStretch()
        A.apply_button = QtWidgets.QPushButton('Apply', A)
        A.apply_button.clicked.connect(A._show_preview)
        D.addWidget(A.apply_button)
        C.addLayout(D)
        A.layout.addWidget(A.preview)
        A.layout.addWidget(A.editor_container)
        A.preview_size = A.preview.sizeHint()
        A.editor_size = QtCore.QSize(120, 80)
        A.layout.currentChanged.connect(A._handle_view_change)
        A.layout.setCurrentIndex(0)
        A.updateGeometry()
        A.editor.curve_changed.connect(A._on_curve_changed)
        A.updating = False

    def _show_editor(A):
        'Switch to editor view'
        A.layout.setCurrentIndex(1)

    def _show_preview(A):
        'Switch to preview view'
        A.layout.setCurrentIndex(0)

    def _handle_view_change(A, index):
        'Handle view changes to adjust size'
        if (index == 0):
            A.setMinimumSize(A.preview.minimumSize())
            A.setMaximumSize(A.preview.maximumSize())
        else:
            A.setMinimumSize(A.editor_size)
            A.setMaximumSize(16777215, 16777215)
        A.updateGeometry()
        A.adjustSize()
        if (A.parent() and isinstance(A.parent().layout(), QtWidgets.QLayout)):
            A.parent().layout().update()

    def _on_curve_changed(A, points):
        'Handle curve changes from the editor'
        B = points
        if A.updating:
            return
        A.preview.set_points(B)
        A.valueChanged.emit(B)

    def value(A):
        'Get the current points'
        return A.editor.points

    def setValue(A, value):
        'Set the points'
        B = value
        A.updating = True
        A.editor.set_points(B)
        A.preview.set_points(B)
        A.valueChanged.emit(B)
        A.updating = False

    def sizeHint(A):
        'Return a size hint based on current view'
        if (A.layout.currentIndex() == 0):
            return A.preview_size
        else:
            return A.editor_size

    def minimumSizeHint(A):
        'Return a minimum size hint based on current view'
        if (A.layout.currentIndex() == 0):
            return A.preview.minimumSize()
        else:
            return QtCore.QSize(200, 200)

class CurveEditorWindow(QtWidgets.QWidget):
    'Main curve editor window'
    curve_changed = QtCore.Signal(list)

    def __init__(A, points, min_value=None, max_value=None, parent=None):
        super().__init__(parent, QtCore.Qt.Window)
        A.setWindowTitle('Curve Editor')
        A.set_points(points)
        A.selected_point = None
        A.min_value = min_value
        A.max_value = max_value
        A.zoom = 1.0
        A.offset = QtCore.QPointF(0, 0)
        C = QtWidgets.QVBoxLayout(A)
        A.toolbar = QtWidgets.QWidget()
        B = QtWidgets.QHBoxLayout(A.toolbar)
        B.setContentsMargins(0, 0, 0, 0)
        A.pos_spin = QtWidgets.QDoubleSpinBox()
        A.pos_spin.setRange(0, 1)
        A.pos_spin.setSingleStep(0.1)
        A.pos_spin.setDecimals(3)
        A.pos_spin.valueChanged.connect(A._on_pos_changed)
        B.addWidget(QtWidgets.QLabel('Position:'))
        B.addWidget(A.pos_spin)
        A.value_spin = QtWidgets.QDoubleSpinBox()
        A.value_spin.setRange((- 9999), 9999)
        A.value_spin.setSingleStep(0.1)
        A.value_spin.setDecimals(3)
        A.value_spin.valueChanged.connect(A._on_value_changed)
        B.addWidget(QtWidgets.QLabel('Value:'))
        B.addWidget(A.value_spin)
        A.interp_combo = QtWidgets.QComboBox()
        A.interp_combo.addItems(CurvePoint.interp_types)
        A.interp_combo.currentIndexChanged.connect(A._on_interp_changed)
        B.addWidget(QtWidgets.QLabel('Interpolation:'))
        B.addWidget(A.interp_combo)
        B.addStretch()
        C.addWidget(A.toolbar)
        A.view = QtWidgets.QWidget()
        A.view.setMinimumSize(120, 80)
        A.view.paintEvent = A._paint_view
        A.view.mousePressEvent = A._view_mouse_press
        A.view.mouseMoveEvent = A._view_mouse_move
        A.view.mouseDoubleClickEvent = A._view_double_click
        A.view.wheelEvent = A._view_wheel
        C.addWidget(A.view)
        A.toolbar.setVisible(False)
        A.resize(600, 400)
        A.point_menu = QtWidgets.QMenu()
        E = A.point_menu.addMenu('Interpolation')
        for D in CurvePoint.interp_types:
            F = E.addAction(D)
            F.setData(('interpolation', D))
        A.point_menu.addSeparator()
        G = A.point_menu.addAction('Delete')
        G.setData(('delete', None))
        A.point_menu.triggered.connect(A._on_context_menu)

    def set_points(C, points):
        B = []
        for A in points:
            if isinstance(A, CurvePoint):
                B.append(A)
            elif (isinstance(A, (list, tuple)) and (len(A) == 3)):
                B.append(CurvePoint(*A))
            else:
                raise ValueError('Invalid point format')
        C.points = sorted(B, key=(lambda x: x.position))

    def _view_to_curve_space(A, pos):
        'Convert view coordinates to curve space'
        B = pos
        (G, D) = (A.view.width(), A.view.height())
        H = ((B.x() - A.offset.x()) / (G * A.zoom))
        if ((A.min_value is not None) and (A.max_value is not None)):
            C = (A.max_value - A.min_value)
            E = (A.max_value - (((B.y() - A.offset.y()) / (D * A.zoom)) * C))
        else:
            I = min((A.value for A in A.points))
            F = max((A.value for A in A.points))
            C = (F - I)
            E = (F - (((B.y() - A.offset.y()) / (D * A.zoom)) * C))
        return (H, E)

    def _curve_to_view_space(A, x, y):
        'Convert curve space coordinates to view space'
        (F, C) = (A.view.width(), A.view.height())
        G = (((x * F) * A.zoom) + A.offset.x())
        if ((A.min_value is not None) and (A.max_value is not None)):
            B = (A.max_value - A.min_value)
            if (B == 0):
                D = (((C * 0.5) * A.zoom) + A.offset.y())
            else:
                D = (((((A.max_value - y) / B) * C) * A.zoom) + A.offset.y())
        else:
            H = min((A.value for A in A.points))
            E = max((A.value for A in A.points))
            B = (E - H)
            if (B == 0):
                D = (((C * 0.5) * A.zoom) + A.offset.y())
            else:
                D = (((((E - y) / B) * C) * A.zoom) + A.offset.y())
        return QtCore.QPointF(G, D)

    def _find_point_at(C, pos, threshold=10):
        'Find a point near the given position'
        A = pos
        A = QtCore.QPointF(A)
        for B in C.points:
            D = C._curve_to_view_space(B.position, B.value)
            if ((A - D).manhattanLength() < threshold):
                return B

    def _view_mouse_press(A, event):
        C = event
        if (C.button() == QtCore.Qt.LeftButton):
            B = A._find_point_at(C.pos())
            A.selected_point = B
            A.toolbar.setVisible(bool(B))
            if B:
                A.pos_spin.setValue(B.position)
                A.value_spin.setValue(B.value)
                A.interp_combo.setCurrentIndex(B.interpolation)
            A.view.update()
        elif (C.button() == QtCore.Qt.RightButton):
            B = A._find_point_at(C.pos())
            if B:
                A.selected_point = B
                A.point_menu.popup(A.view.mapToGlobal(C.pos()))

    def _view_mouse_move(A, event):
        C = event
        if ((C.buttons() & QtCore.Qt.LeftButton) and A.selected_point):
            (B, D) = A._view_to_curve_space(C.pos())
            B = max(0, min(1, B))
            A.selected_point.position = B
            A.selected_point.value = D
            A.pos_spin.setValue(B)
            A.value_spin.setValue(D)
            A.view.update()
            A.curve_changed.emit(A.points)

    def _view_double_click(A, event):
        (B, C) = A._view_to_curve_space(event.pos())
        if (0 <= B <= 1):
            D = CurvePoint(B, C)
            A.points.append(D)
            A.points.sort(key=(lambda p: p.position))
            A.selected_point = D
            A.toolbar.setVisible(True)
            A.pos_spin.setValue(B)
            A.value_spin.setValue(C)
            A.view.update()
            A.curve_changed.emit(A.points)

    def _view_wheel(A, event):
        B = (1.2 if (event.angleDelta().y() > 0) else (1 / 1.2))
        A.zoom *= B
        A.view.update()

    def _paint_view(A, event):
        C = QtGui.QPainter(A.view)
        C.setRenderHint(QtGui.QPainter.Antialiasing)
        C.fillRect(A.view.rect(), QtGui.QColor(30, 30, 30))
        I = QtGui.QPen(QtGui.QColor(50, 50, 50))
        C.setPen(I)
        for D in range(11):
            T = (D / 10)
            F = A._curve_to_view_space(T, 0)
            C.drawLine(F.x(), 0, F.x(), A.view.height())
        if ((A.min_value is not None) and (A.max_value is not None)):
            U = (A.max_value - A.min_value)
            V = (U / 10)
            for D in range(11):
                W = (A.min_value + (D * V))
                F = A._curve_to_view_space(0, W)
                C.drawLine(0, F.y(), A.view.width(), F.y())
        if A.points:
            I = QtGui.QPen(QtGui.QColor(201, 145, 38))
            I.setWidth(2)
            C.setPen(I)
            G = QtGui.QPainterPath()
            N = A.points[0]
            O = A._curve_to_view_space(N.position, N.value)
            G.moveTo(O)
            for D in range((len(A.points) - 1)):
                H = A.points[D]
                P = A.points[(D + 1)]
                E = A._curve_to_view_space(H.position, H.value)
                B = A._curve_to_view_space(P.position, P.value)
                if (H.interpolation == CurvePoint.LINEAR):
                    G.lineTo(B)
                elif (H.interpolation == CurvePoint.SMOOTH):
                    K = (E + QtCore.QPointF(((B.x() - E.x()) * 0.5), 0))
                    L = (B - QtCore.QPointF(((B.x() - E.x()) * 0.5), 0))
                    G.cubicTo(K, L, B)
                elif (H.interpolation == CurvePoint.SPLINE):
                    J = 0.5
                    Q = (B.x() - E.x())
                    b = (B.y() - E.y())
                    R = E
                    if (D > 0):
                        R = A._curve_to_view_space(A.points[(D - 1)].position, A.points[(D - 1)].value)
                    S = B
                    if (D < (len(A.points) - 2)):
                        S = A._curve_to_view_space(A.points[(D + 2)].position, A.points[(D + 2)].value)
                    X = (E.x() + (Q * J))
                    Y = (E.y() + (((B.y() - R.y()) * J) * 0.5))
                    Z = (B.x() - (Q * J))
                    a = (B.y() - (((S.y() - E.y()) * J) * 0.5))
                    K = QtCore.QPointF(X, Y)
                    L = QtCore.QPointF(Z, a)
                    G.cubicTo(K, L, B)
                else:
                    G.lineTo(QtCore.QPointF(B.x(), E.y()))
                    G.lineTo(B)
                O = B
            C.drawPath(G)
            for M in A.points:
                F = A._curve_to_view_space(M.position, M.value)
                if (M == A.selected_point):
                    C.setBrush(QtGui.QBrush(QtGui.QColor(255, 255, 0)))
                    C.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0)))
                    C.drawEllipse(F, 6, 6)
                else:
                    C.setBrush(QtGui.QBrush(QtGui.QColor(201, 145, 38)))
                    C.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0)))
                    C.drawEllipse(F, 4, 4)

    def _on_pos_changed(A, value):
        if A.selected_point:
            A.selected_point.position = value
            A.points.sort(key=(lambda p: p.position))
            A.view.update()
            A.curve_changed.emit(A.points)

    def _on_value_changed(A, value):
        if A.selected_point:
            A.selected_point.value = value
            A.view.update()
            A.curve_changed.emit(A.points)

    def _on_interp_changed(A, index):
        if A.selected_point:
            A.selected_point.interpolation = index
            A.view.update()
            A.curve_changed.emit(A.points)

    def _on_context_menu(A, action):
        if A.selected_point:
            (B, C) = action.data()
            if (B == 'interpolation'):
                A.selected_point.interpolation = CurvePoint.interp_types.index(C)
                A.interp_combo.setCurrentIndex(A.selected_point.interpolation)
            elif ((B == 'delete') and (len(A.points) > 2)):
                A.points.remove(A.selected_point)
                A.selected_point = None
                A.toolbar.setVisible(False)
            A.view.update()
            A.curve_changed.emit(A.points)
