from base64 import b64decode as license64

from BroTools.core.broqt.vendor.Qt import QtCore, QtGui, QtWidgets
import uuid, json
from BroTools.core.debug import default_logger as log
from BroTools.core.errors import catch_errors
from BroTools.core.broqt.color import ColorPalette
from dataclasses import dataclass, asdict
from typing import Any, Dict, List, Type, Optional
from BroTools.BroAI.ai_core.utils.templating import render_template_string
from BroTools.core.broqt.widgets.curve_editor import CurvePreviewWidget, CurveEditorWindow
from BroTools.core.curves import CurvePoint

@dataclass
class DCCObjectReference():
    'Base class for operations with DCCs like Maya and Blender that will be later overriden.\n    This class is serializable and can be converted to/from JSON format using json.dumps/loads.\n    '
    node_id: str

    def __post_init__(self):
        pass

    def __json__(self):
        'Make the class JSON serializable by converting to a dictionary.\n        This method is called by JSON encoders when serializing the object.\n        '
        data = asdict(self)
        data['__class__'] = self.__class__.__name__
        data['__module__'] = self.__class__.__module__
        return data

    @classmethod
    def from_dict(cls, data):
        'Create an instance from a dictionary.\n        \n        Args:\n            data: Dictionary containing the object data\n            \n        Returns:\n            A new instance of the class\n        '
        if (('__class__' in data) and ('__module__' in data)):
            class_name = data.pop('__class__')
            module_name = data.pop('__module__')
            try:
                import importlib
                module = importlib.import_module(module_name)
                target_cls = getattr(module, class_name)
                return target_cls(**data)
            except (ImportError, AttributeError) as e:
                print(f'Warning: Could not import {module_name}.{class_name}, falling back to base class')
        return cls(**data)

    @property
    def name(self):
        'Returns the short name of the object.\n        This is a base implementation that will be overridden by specific DCC implementations.\n        '
        return ''

    @property
    def full_name(self):
        'Returns the full name/path of the object.\n        This is a base implementation that will be overridden by specific DCC implementations.\n        '
        return ''

class NodeAttribute():
    'Base class for node attributes'

    def __init__(self, default_value=None, label=None, tooltip=None):
        self.default_value = default_value
        self.label = label
        self.tooltip = tooltip
        self.widget = None
        self._value = default_value

    def create_widget(self):
        'Create the Qt widget for this attribute'
        raise NotImplementedError()

    @property
    def value(self):
        'Get the current value'
        return self._value

    @value.setter
    def value(self, new_value):
        'Set a new value'
        self._value = new_value
        if self.widget:
            self._update_widget()

    def _update_widget(self):
        'Update the widget to reflect the current value'
        raise NotImplementedError()

    def serialize(self):
        'Serialize the attribute to a dictionary'
        return {'default_value': self.default_value, 'label': self.label, 'tooltip': self.tooltip, 'value': self._value}

    @classmethod
    def deserialize(cls, data):
        'Create an instance from serialized data'
        instance = cls()
        for (key, value) in data.items():
            if ((not key.startswith('_')) and hasattr(instance, key)):
                setattr(instance, key, value)
        return instance

class FloatAttribute(NodeAttribute):
    'Float input attribute'

    def __init__(self, default_value=0.0, min_value=None, max_value=None, **kwargs):
        super().__init__(default_value=default_value, **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def create_widget(self):
        self.widget = QtWidgets.QDoubleSpinBox()
        if (self.min_value is not None):
            self.widget.setMinimum(self.min_value)
        if (self.max_value is not None):
            self.widget.setMaximum(self.max_value)
        self.widget.setValue(self._value)
        self.widget.valueChanged.connect((lambda v: setattr(self, 'value', v)))
        if self.tooltip:
            self.widget.setToolTip(self.tooltip)
        self.widget.setStyleSheet('\n            QDoubleSpinBox {\n                background-color: transparent;\n                border: 1px solid #666666;\n                border-radius: 2px;\n                color: #FFFFFF;\n                padding: 2px;\n            }\n            QDoubleSpinBox:hover {\n                border: 1px solid #999999;\n            }\n            QDoubleSpinBox::up-arrow:hover, QDoubleSpinBox::down-arrow:hover {\n                border-bottom-color: #AAAAAA;\n                border-top-color: #AAAAAA;\n            }\n        ')
        return self.widget

    def _update_widget(self):
        if isinstance(self.widget, QtWidgets.QDoubleSpinBox):
            self.widget.setValue(self._value)

class IntAttribute(NodeAttribute):
    'Integer input attribute'

    def __init__(self, default_value=0, min_value=None, max_value=None, **kwargs):
        super().__init__(default_value=default_value, **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def create_widget(self):
        self.widget = QtWidgets.QSpinBox()
        if (self.min_value is not None):
            self.widget.setMinimum(self.min_value)
        if (self.max_value is not None):
            self.widget.setMaximum(self.max_value)
        self.widget.setValue(self._value)
        self.widget.valueChanged.connect((lambda v: setattr(self, 'value', v)))
        if self.tooltip:
            self.widget.setToolTip(self.tooltip)
        self.widget.setStyleSheet('\n            QSpinBox {\n                background-color: transparent;\n                border: 1px solid #666666;\n                border-radius: 2px;\n                color: #FFFFFF;\n                padding: 2px;\n            }\n            QSpinBox:hover {\n                border: 1px solid #999999;\n            }\n            QSpinBox::up-arrow:hover, QSpinBox::down-arrow:hover {\n                border-bottom-color: #AAAAAA;\n                border-top-color: #AAAAAA;\n            }\n        ')
        return self.widget

    def _update_widget(self):
        if isinstance(self.widget, QtWidgets.QSpinBox):
            self.widget.setValue(self._value)

class BoolAttribute(NodeAttribute):
    'Boolean input attribute'

    def __init__(self, default_value=False, **kwargs):
        super().__init__(default_value=default_value, **kwargs)

    def create_widget(self):
        self.widget = QtWidgets.QCheckBox()
        self.widget.setChecked(self._value)
        self.widget.stateChanged.connect((lambda v: setattr(self, 'value', bool(v))))
        if self.tooltip:
            self.widget.setToolTip(self.tooltip)
        self.widget.setStyleSheet(render_template_string('\n            QCheckBox {\n                color: #FFFFFF;\n                spacing: 5px;\n            }\n            QCheckBox::indicator {\n                width: 13px;\n                height: 13px;\n                background-color: transparent;\n                border: 1px solid {{bg}};\n                border-radius: 2px;\n            }\n            QCheckBox::indicator:hover {\n                border: 1px solid {{hover}};\n            }\n            QCheckBox::indicator:checked {\n                background-color: {{checked}};\n            }\n            QCheckBox::indicator:checked:hover {\n                background-color: {{checked_hover}};\n            }\n        ', dict(bg=ColorPalette.instance()._base_palette['gray']['400'], hover=ColorPalette.instance()._base_palette['gray']['100'], checked=ColorPalette.instance()._base_palette['green']['500'], checked_hover=ColorPalette.instance()._base_palette['green']['200'])))
        return self.widget

    def _update_widget(self):
        if isinstance(self.widget, QtWidgets.QCheckBox):
            self.widget.setChecked(self._value)

class CurveAttribute(NodeAttribute):
    'Curve attribute that allows editing an array of curve points with position, value and interpolation type'

    def __init__(self, default_value=None, min_value=None, max_value=None, **kwargs):
        if (default_value is None):
            default_value = [CurvePoint(0.0, 0.0, CurvePoint.LINEAR), CurvePoint(1.0, 1.0, CurvePoint.LINEAR)]
        elif isinstance(default_value, list):
            default_value = [CurvePoint(p.position, p.value, p.interpolation) for p in default_value]
        super().__init__(default_value=default_value, **kwargs)
        self.min_value = min_value
        self.max_value = max_value
        self.editor_window = None

    def create_widget(self):
        'Create a preview widget that opens the curve editor when clicked'
        container = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(container)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)
        if self.label:
            label = QtWidgets.QLabel(self.label)
            layout.addWidget(label)
        self.widget = CurvePreviewWidget()
        self.widget.clicked.connect(self._open_editor)
        layout.addWidget(self.widget)
        layout.addStretch()
        if self.tooltip:
            container.setToolTip(self.tooltip)
        self._update_widget()
        return container

    def _update_widget(self):
        'Update the preview widget with current curve points'
        if self.widget:
            self.widget.set_points(self._value)

    def _open_editor(self):
        'Open the curve editor window'
        if (not self.editor_window):
            self.editor_window = CurveEditorWindow(self._value, min_value=self.min_value, max_value=self.max_value)
            self.editor_window.curve_changed.connect(self._on_curve_changed)
        self.editor_window.show()
        self.editor_window.raise_()

    def _on_curve_changed(self, points):
        'Called when curve is modified in the editor'
        self.value = points

    def serialize(self):
        'Serialize the curve points'
        data = super().serialize()
        data['points'] = [point.__json__() for point in self._value]
        if (self.min_value is not None):
            data['min_value'] = self.min_value
        if (self.max_value is not None):
            data['max_value'] = self.max_value
        return data

    @classmethod
    def deserialize(cls, data):
        'Deserialize curve points from saved data'
        instance = cls()
        if ('min_value' in data):
            instance.min_value = data['min_value']
        if ('max_value' in data):
            instance.max_value = data['max_value']
        if ('label' in data):
            instance.label = data['label']
        if ('tooltip' in data):
            instance.tooltip = data['tooltip']
        if ('points' in data):
            points = [CurvePoint.from_dict(point_data) for point_data in data['points']]
            instance._value = points
        elif (('value' in data) and isinstance(data['value'], list)):
            points = []
            for point_data in data['value']:
                if isinstance(point_data, dict):
                    points.append(CurvePoint.from_dict(point_data))
                elif isinstance(point_data, CurvePoint):
                    points.append(CurvePoint(point_data.position, point_data.value, point_data.interpolation))
            if points:
                instance._value = points
        else:
            instance._value = instance.default_value
        return instance

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        'Set new curve points and update the widget if it exists'
        if (not isinstance(new_value, list)):
            raise TypeError('Value must be a list of CurvePoint objects')
        if (not all((isinstance(p, CurvePoint) for p in new_value))):
            raise TypeError('All points must be CurvePoint objects')
        if (len(new_value) < 2):
            raise ValueError('Must have at least 2 points')
        self._value = sorted(new_value, key=(lambda p: p.position))
        if self.widget:
            self._update_widget()

class NodeSocket(QtWidgets.QGraphicsItem):
    'Represents an input or output connection point on a node'

    def __init__(self, parent=None, is_output=False, socket_type='any', is_array=False, ordered_array=False):
        super().__init__(parent)
        self.is_output = is_output
        self.socket_type = socket_type
        self.is_array = is_array
        self.ordered_array = ordered_array
        self.connections = []
        self._radius = 6
        self._hover = False
        self._vertical_spacing = 20
        self.sub_sockets = []
        self._connection_to_socket = {}
        if ordered_array:
            self._add_sub_socket()
        self.setAcceptHoverEvents(True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsScenePositionChanges)
        self.setAcceptHoverEvents(True)

    def _add_sub_socket(self):
        'Add a new visual sub-socket for ordered arrays'
        socket_y = (len(self.sub_sockets) * self._vertical_spacing)
        sub_socket = {'pos': QtCore.QPointF(0, socket_y), 'connection': None}
        self.sub_sockets.append(sub_socket)
        if self.parentItem():
            QtCore.QTimer.singleShot(0.1, self.parentItem().updateGeometry)
        return (len(self.sub_sockets) - 1)

    def _remove_sub_socket(self, index):
        'Remove a sub-socket and shift others up'
        if (0 <= index < len(self.sub_sockets)):
            self.sub_sockets.pop(index)
            self._rearrange_connections()

    def _rearrange_connections(self):
        'Rearrange connections to remove gaps and cleanup empty sockets'
        if getattr(self, '_is_rearranging', False):
            return
        self._is_rearranging = True
        try:
            if (not self.ordered_array):
                return
            new_sockets = []
            new_mapping = {}
            active_connections = []
            for (i, sub_socket) in enumerate(self.sub_sockets):
                if (sub_socket['connection'] in self.connections):
                    active_connections.append(sub_socket['connection'])
            for (i, connection) in enumerate(active_connections):
                pos = QtCore.QPointF(0, (i * self._vertical_spacing))
                new_sockets.append({'pos': pos, 'connection': connection})
                new_mapping[connection] = i
            if active_connections:
                pos = QtCore.QPointF(0, (len(active_connections) * self._vertical_spacing))
                new_sockets.append({'pos': pos, 'connection': None})
            else:
                new_sockets.append({'pos': QtCore.QPointF(0, 0), 'connection': None})
            self.sub_sockets = new_sockets
            self._connection_to_socket = new_mapping
        finally:
            self._is_rearranging = False
            try:
                if (self.parentItem() and (not getattr(self, '_pending_update', False))):
                    self._pending_update = True
                    QtCore.QTimer.singleShot(0.1, (lambda : (self.parentItem().updateGeometry(), setattr(self, '_pending_update', False))))
            except RuntimeError as e:
                if ('Internal C++ object' in e):
                    log.debug(f'Expected error, all good: {e}')
                else:
                    log.warning(f'Unexpect error, but should be good to proceed: {e}', exc_info=True)
                pass

    def removeConnection(self, connection):
        'Remove a connection from this socket'
        try:
            if (connection in self.connections):
                self.connections.remove(connection)
                if self.ordered_array:
                    if (connection in self._connection_to_socket):
                        idx = self._connection_to_socket[connection]
                        if (0 <= idx < len(self.sub_sockets)):
                            self.sub_sockets[idx]['connection'] = None
                        del self._connection_to_socket[connection]
                    QtCore.QTimer.singleShot(0.1, (lambda : self._rearrange_connections()))
        except Exception as e:
            log.warning(f'Error removing connection: {e}')

    def removeAllConnections(self):
        'Remove all connections from this socket'
        try:
            connections = self.connections.copy()
            for conn in connections:
                if (conn in self.connections):
                    conn.remove()
        except Exception as e:
            log.warning(f'Error removing all connections: {e}')
            self.connections = []
            self._connection_to_socket = {}

    def addConnection(self, connection):
        'Add a connection to this socket'
        if ((not self.is_array) and (not self.is_output) and self.connections):
            return False
        if self.ordered_array:
            for (i, sub_socket) in enumerate(self.sub_sockets):
                if (sub_socket['connection'] is None):
                    sub_socket['connection'] = connection
                    self._connection_to_socket[connection] = i
                    self.connections.append(connection)
                    if (i == (len(self.sub_sockets) - 1)):
                        self._add_sub_socket()
                    return True
        else:
            if self.is_output:
                for existing_conn in self.connections:
                    if (existing_conn.end_socket == connection.end_socket):
                        return False
            self.connections.append(connection)
            if (self.is_array and (not self.is_output)):
                self._connection_to_socket[connection] = (len(self.connections) - 1)
            return True
        return False

    def getSubSocketPos(self, connection):
        "Get the scene position for a specific connection's sub-socket"
        if (self.ordered_array and (connection in self._connection_to_socket)):
            idx = self._connection_to_socket[connection]
            if (0 <= idx < len(self.sub_sockets)):
                return self.mapToScene(self.sub_sockets[idx]['pos'])
        return self.scenePos()

    def boundingRect(self):
        'Return bounding rectangle, accounting for sub-sockets'
        if self.ordered_array:
            height = (len(self.sub_sockets) * self._vertical_spacing)
            return QtCore.QRectF((- self._radius), (- self._radius), (self._radius * 2), (height + (self._radius * 2)))
        return QtCore.QRectF((- self._radius), (- self._radius), (self._radius * 2), (self._radius * 2))

    def paint(self, painter, option, widget):
        'Paint the socket(s)'
        if self.ordered_array:
            for sub_socket in self.sub_sockets:
                painter.save()
                painter.translate(sub_socket['pos'])
                if sub_socket['connection']:
                    painter.setBrush(QtGui.QBrush(ColorPalette.instance().SOCKET_CONNECTED_BACKGROUND))
                    painter.setPen(QtGui.QPen(QtGui.QColor(ColorPalette.instance().SOCKET_CONNECTED_BACKGROUND).darker(240), 1.5))
                else:
                    painter.setBrush(QtGui.QBrush(QtGui.QColor(ColorPalette.instance().SOCKET_BACKGROUND)))
                    painter.setPen(QtGui.QPen(QtGui.QColor(ColorPalette.instance().SOCKET_BACKGROUND).darker(240), 1.5))
                painter.drawEllipse(QtCore.QRectF((- self._radius), (- self._radius), (self._radius * 2), (self._radius * 2)))
                painter.restore()
        else:
            if self.connections:
                painter.setBrush(QtGui.QBrush(QtGui.QColor(ColorPalette.instance().SOCKET_CONNECTED_BACKGROUND)))
                painter.setPen(QtGui.QPen(QtGui.QColor(ColorPalette.instance().SOCKET_CONNECTED_BACKGROUND).darker(240), 1.5))
            else:
                painter.setBrush(QtGui.QBrush(QtGui.QColor(ColorPalette.instance().SOCKET_BACKGROUND)))
                painter.setPen(QtGui.QPen(QtGui.QColor(ColorPalette.instance().SOCKET_BACKGROUND).darker(240), 1.5))
            painter.drawEllipse(self.boundingRect())

    def scenePos(self):
        'Return position for connection lines'
        if (self.ordered_array and self.parentItem()):
            for (i, sub_socket) in enumerate(self.sub_sockets):
                if self.is_output:
                    for conn in self.connections:
                        if (conn.start_socket == self):
                            return self.mapToScene(sub_socket['pos'])
                elif (sub_socket['connection'] in self.connections):
                    return self.mapToScene(sub_socket['pos'])
        return super().scenePos()

    def itemChange(self, change, value):
        'Handle position changes to update connected edges'
        if (change == QtWidgets.QGraphicsItem.ItemScenePositionHasChanged):
            for conn in self.connections:
                conn.updatePath()
        return super().itemChange(change, value)

    def hoverEnterEvent(self, event):
        self._hover = True
        self.update()
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self._hover = False
        self.update()
        super().hoverLeaveEvent(event)

    def mousePressEvent(self, event):
        'Handle mouse press to start connection'
        if (event.button() == QtCore.Qt.LeftButton):
            if self.scene():
                self.scene().startConnection(self)
        super().mousePressEvent(event)

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu()
        if self.connections:
            remove_all = menu.addAction('Remove All Connections')
            remove_all.triggered.connect(self.removeAllConnections)
        if (not menu.isEmpty()):
            menu.exec_(event.screenPos())

    def serialize(self):
        'Serialize socket data'
        data = {'is_output': self.is_output, 'socket_type': self.socket_type, 'is_array': self.is_array, 'ordered_array': self.ordered_array, 'index': (self.parentItem().input_sockets.index(self) if (not self.is_output) else self.parentItem().output_sockets.index(self))}
        if self.ordered_array:
            connection_order = []
            for sub_socket in self.sub_sockets:
                if (sub_socket['connection'] in self.connections):
                    connection_order.append(self.connections.index(sub_socket['connection']))
            data['connection_order'] = connection_order
        return data

    def can_connect_to(self, other_socket):
        'Check if this socket can connect to another socket'
        if (self.is_output == other_socket.is_output):
            return False
        if (self.parentItem() == other_socket.parentItem()):
            return False
        if self.is_output:
            output_socket = self
            input_socket = other_socket
        else:
            output_socket = other_socket
            input_socket = self
        if ((not input_socket.is_array) and input_socket.connections):
            return False
        for conn in output_socket.connections:
            if (conn.end_socket == input_socket):
                return False
        return True

class Node(QtWidgets.QGraphicsItem):
    'Base class for all nodes in the system'
    WIDTH = 160
    HEIGHT = 80
    TITLE_HEIGHT = 24
    registry = {}
    attributes = {}

    def __init__(self, title=None, parent=None):
        'Initialize a node'
        super().__init__(parent)
        self.title = (title or self.__class__.__name__)
        self.width = self.WIDTH
        self.height = self.HEIGHT
        self.input_sockets = []
        self.output_sockets = []
        self.id = str(uuid.uuid4())
        self.dcc_object = None
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable)
        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsScenePositionChanges)
        self._is_selected = False
        self._attributes = {}
        self._attribute_widgets = {}
        self._init_attributes()
        self._init_ui()

    def _init_ui(self):
        "Initialize the node's visual elements"
        self._attribute_container = None
        if self.attributes:
            self._attribute_container = QtWidgets.QGraphicsProxyWidget(self)
            widget = self.create_attribute_widgets()
            self._attribute_container.setWidget(widget)
            self._attribute_container.setPos(5, (self.TITLE_HEIGHT + 5))

    def _init_attributes(self):
        'Initialize instance attributes from class definitions'
        for (name, attr_def) in self.attributes.items():
            attr_instance = type(attr_def)()
            for (key, value) in vars(attr_def).items():
                if (not key.startswith('_')):
                    if ((key == 'default_value') and (value is not None)):
                        attr_instance.default_value = value
                        attr_instance._value = value
                    else:
                        setattr(attr_instance, key, value)
            self._attributes[name] = attr_instance

    def create_attribute_widgets(self):
        'Create a widget containing all attribute editors'
        container = QtWidgets.QWidget()
        container.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        layout = QtWidgets.QFormLayout(container)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(4)
        for (name, attr) in self._attributes.items():
            widget = attr.create_widget()
            if widget:
                label = QtWidgets.QLabel((attr.label or name.replace('_', ' ').title()))
                label.setStyleSheet('color: #CCCCCC; font-size: 10px;')
                layout.addRow(label, widget)
                self._attribute_widgets[name] = widget
        return container

    def get_attribute(self, name):
        'Get the value of an attribute'
        if (name not in self._attributes):
            raise KeyError(f'Attribute {name} not found')
        return self._attributes[name].value

    def set_attribute(self, name, value):
        'Set the value of an attribute'
        if (name not in self._attributes):
            raise KeyError(f'Attribute {name} not found')
        self._attributes[name].value = value

    def serialize(self):
        'Serialize node data including attributes'
        pos = self.pos()
        data = {'id': self.id, 'type': self.__class__.__name__, 'pos': [pos.x(), pos.y()], 'attributes': {}}
        if self.dcc_object:
            if hasattr(self.dcc_object, '__json__'):
                data['dcc_object'] = self.dcc_object.__json__()
            else:
                data['dcc_object'] = vars(self.dcc_object)
        for (name, attr) in self._attributes.items():
            data['attributes'][name] = attr.serialize()
        return data

    @classmethod
    def deserialize(cls, data, scene=None):
        'Create a node from serialized data'
        node_class = cls.registry.get(data['type'])
        if (not node_class):
            raise ValueError(f"Unknown node type: {data['type']}")
        try:
            node = node_class(title=data.get('title'))
        except TypeError:
            node = node_class()
        node.id = data['id']
        if ('pos' in data):
            node.setPos(QtCore.QPointF(data['pos'][0], data['pos'][1]))
        else:
            node.setPos(QtCore.QPointF(data.get('x', 0), data.get('y', 0)))
        if ('dcc_object' in data):
            dcc_data = data['dcc_object']
            if isinstance(dcc_data, dict):
                node.dcc_object = DCCObjectReference.from_dict(dcc_data)
            else:
                node.dcc_object = dcc_data
        if ('attributes' in data):
            node._attributes.clear()
            for (attr_name, attr_data) in data['attributes'].items():
                if (attr_name in node.attributes):
                    attr_type = type(node.attributes[attr_name])
                    attr_instance = attr_type.deserialize(attr_data)
                    node._attributes[attr_name] = attr_instance
        node._init_ui()
        if scene:
            scene.addItem(node)
        return node

    @classmethod
    def register(cls, node_class):
        'Register a node class in the registry'
        cls.registry[node_class.__name__] = node_class
        return node_class

    def addInputSocket(self, socket_type='any', is_array=False, ordered_array=False):
        'Add an input socket to the node'
        socket = NodeSocket(self, is_output=False, socket_type=socket_type, is_array=is_array, ordered_array=ordered_array)
        self.input_sockets.append(socket)
        self.updateGeometry()
        return socket

    def addOutputSocket(self, socket_type='any', is_array=False, ordered_array=False):
        'Add an output socket to the node'
        socket = NodeSocket(self, is_output=True, socket_type=socket_type, is_array=is_array, ordered_array=ordered_array)
        self.output_sockets.append(socket)
        self.updateGeometry()
        return socket

    def updateGeometry(self):
        'Update node geometry to accommodate sockets'
        input_height = 0
        output_height = 0
        attribute_height = 0
        for socket in self.input_sockets:
            if socket.ordered_array:
                socket_height = (len(socket.sub_sockets) * socket._vertical_spacing)
                input_height = max(input_height, socket_height)
            else:
                input_height = max(input_height, (socket._radius * 2))
        for socket in self.output_sockets:
            if socket.ordered_array:
                socket_height = (len(socket.sub_sockets) * socket._vertical_spacing)
                output_height = max(output_height, socket_height)
            else:
                output_height = max(output_height, (socket._radius * 2))
        if (self._attribute_container and self._attribute_container.widget()):
            attribute_height = self._attribute_container.widget().sizeHint().height()
        self.height = max(self.HEIGHT, (((self.TITLE_HEIGHT + attribute_height) + max(input_height, output_height)) + 30))
        self.updateSockets()
        if (self._attribute_container and self._attribute_container.widget()):
            self._attribute_container.widget().setFixedWidth((self.width - 10))

    def updateSockets(self):
        'Update the positions of all sockets'
        for (i, socket) in enumerate(self.input_sockets):
            if socket.ordered_array:
                socket.setPos(0, (self.TITLE_HEIGHT + 20))
            else:
                y = ((self.TITLE_HEIGHT + 20) + (i * 20))
                socket.setPos(0, y)
        for (i, socket) in enumerate(self.output_sockets):
            if socket.ordered_array:
                socket.setPos(self.width, (self.TITLE_HEIGHT + 20))
            else:
                y = ((self.TITLE_HEIGHT + 20) + (i * 20))
                socket.setPos(self.width, y)

    def paint(self, painter, option, widget):
        if (not self.HEADER_COLORS):
            raise Exception('HEADER_COLORS not set, when using node_editor your must subclass Node and define a HEADER_COLORS class attribute, which is a dict of colors named by node class')
        header_color = QtGui.QColor(self.HEADER_COLORS.get(self.__class__.__name__, ColorPalette.instance().NODE_DEFAULT_HEADER))
        background_color = QtGui.QColor(ColorPalette.instance().NODE_BACKGROUND)
        border_color = background_color.darker(240)
        border_width = 1
        if self.isSelected():
            background_color = background_color.lighter(120)
            border_color = header_color.darker(240)
            border_width = 2
        painter.setPen(QtGui.QPen(border_color, border_width))
        painter.setBrush(QtGui.QBrush(background_color))
        body_path = QtGui.QPainterPath()
        body_path.addRoundedRect(0, 0, self.width, self.height, 8, 8)
        painter.drawPath(body_path)
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtGui.QBrush(header_color))
        header_rect = QtCore.QRectF(0, 0, self.width, self.TITLE_HEIGHT)
        header_path = QtGui.QPainterPath()
        header_path.addRoundedRect(header_rect, 8, 8)
        header_bottom = QtGui.QPainterPath()
        header_bottom.addRect(0, (self.TITLE_HEIGHT / 2), self.width, (self.TITLE_HEIGHT / 2))
        header_path = header_path.united(header_bottom)
        painter.drawPath(header_path)
        painter.setFont(QtGui.QFont('Arial', 10, QtGui.QFont.Bold))
        text_rect = QtCore.QRectF(0, 0, self.width, self.TITLE_HEIGHT)
        text_flags = ((QtCore.Qt.AlignCenter | QtCore.Qt.AlignVCenter) | QtCore.Qt.TextSingleLine)
        display_text = self.title
        if self.dcc_object:
            try:
                node_name = (self.dcc_object.name if hasattr(self.dcc_object, 'name') else str(self.dcc_object))
                display_text = f'{self.title}: {node_name}'
            except Exception as e:
                log.warning(f'Failed to access DCC object name: {e}')
                self.dcc_object = None
        painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 50)))
        painter.drawText(text_rect.translated(1, 1), text_flags, display_text)
        painter.setPen(QtGui.QPen(QtGui.QColor(ColorPalette.instance().TEXT_LIGHT)))
        painter.drawText(text_rect, text_flags, display_text)

    def mousePressEvent(self, event):
        if (event.button() == QtCore.Qt.LeftButton):
            if (self.scene() and (not (event.modifiers() & QtCore.Qt.ControlModifier))):
                self.scene().clearSelection()
            self.setSelected(True)
            if self.scene():
                self.scene().node_selection_changed()
        super().mousePressEvent(event)

    def boundingRect(self):
        return QtCore.QRectF(0, 0, self.width, self.height)

    def itemChange(self, change, value):
        if (change == QtWidgets.QGraphicsItem.ItemScenePositionHasChanged):
            for socket in (self.input_sockets + self.output_sockets):
                for conn in socket.connections:
                    conn.updatePath()
        return super().itemChange(change, value)

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu()
        delete_action = menu.addAction('Delete Node')
        delete_action.triggered.connect(self.remove)
        menu.exec_(event.screenPos())

    def remove(self):
        'Remove the node and all its connections'
        log.debug(f'Removing node {self}')
        try:
            for socket in (self.input_sockets + self.output_sockets):
                socket.removeAllConnections()
            self.dcc_object = None
            if self.scene():
                self.scene().removeItem(self)
        except Exception as e:
            log.warning(f'Error removing node: {e}')

class NodeScene(QtWidgets.QGraphicsScene):

    def __init__(self):
        super().__init__()
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(ColorPalette.instance().SCENE_BACKGROUND)))
        self.current_connection = None
        self.connecting_socket = None
        self.temp_connection = None
        self.on_selection_changed = None

    def serialize(self):
        'Serialize the entire scene to JSON'
        nodes = [item for item in self.items() if isinstance(item, Node)]
        connections = [item for item in self.items() if isinstance(item, NodeConnection)]
        nodes_data = [node.serialize() for node in nodes]
        array_connections = []
        other_connections = []
        for conn in connections:
            if (isinstance(conn.end_socket, NodeSocket) and conn.end_socket.is_array and (conn in conn.end_socket._connection_to_socket)):
                array_connections.append(conn)
            else:
                other_connections.append(conn)
        array_connections.sort(key=(lambda c: c.end_socket._connection_to_socket[c]))
        connections_data = []
        for conn in (array_connections + other_connections):
            conn_data = conn.serialize()
            if conn_data:
                connections_data.append(conn_data)
        return {'nodes': nodes_data, 'connections': connections_data}

    def deserialize(self, data):
        'Restore the scene from serialized data'
        self.clear()
        nodes = {}
        for node_data in data['nodes']:
            node_class = Node.registry.get(node_data['type'])
            if (not node_class):
                print(f"Warning: Node type {node_data['type']} not found in registry")
                continue
            node = node_class.deserialize(node_data, self)
            nodes[node_data['id']] = node
        for conn_data in data['connections']:
            start_node = nodes.get(conn_data['start_node'])
            end_node = nodes.get(conn_data['end_node'])
            if (not (start_node and end_node)):
                continue
            start_socket = (start_node.output_sockets if (conn_data['start_socket_type'] == 'output') else start_node.input_sockets)[conn_data['start_socket']]
            end_socket = (end_node.output_sockets if (conn_data['end_socket_type'] == 'output') else end_node.input_sockets)[conn_data['end_socket']]
            connection = NodeConnection(start_socket, end_socket)
            self.addItem(connection)
            connection.finalize()
            connection.updatePath()

    def startConnection(self, socket):
        'Start creating a new connection from the given socket'
        self.connecting_socket = socket
        self.temp_connection = NodeConnection(socket)
        self.addItem(self.temp_connection)

    def node_selection_changed(self):
        'Called when node selection changes'
        if self.on_selection_changed:
            selected_nodes = [item for item in self.selectedItems() if isinstance(item, Node)]
            self.on_selection_changed(selected_nodes)

    def mousePressEvent(self, event):
        clicked_item = self.itemAt(event.scenePos(), QtGui.QTransform())
        if (event.button() == QtCore.Qt.LeftButton):
            if (not clicked_item):
                self.clearSelection()
                self.node_selection_changed()
            elif isinstance(clicked_item, NodeSocket):
                self.startConnection(clicked_item)
        elif (event.button() == QtCore.Qt.RightButton):
            event.accept()
            return
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self.temp_connection:
            self.temp_connection.end_socket = event.scenePos()
            self.temp_connection.updatePath()
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self.connecting_socket:
            items = self.items(event.scenePos())
            end_item = None
            for item in items:
                if isinstance(item, NodeSocket):
                    end_item = item
                    break
            if (end_item and self.connecting_socket.can_connect_to(end_item)):
                connection = self.temp_connection
                connection.end_socket = end_item
                if connection.finalize():
                    self.temp_connection = None
                else:
                    self.removeItem(connection)
            else:
                self.removeItem(self.temp_connection)
            self.connecting_socket = None
            self.temp_connection = None
        super().mouseReleaseEvent(event)

class NodeView(QtWidgets.QGraphicsView):

    def __init__(self):
        super().__init__()
        self.setRenderHints(QtGui.QPainter.Antialiasing)
        self.setViewportUpdateMode(QtWidgets.QGraphicsView.FullViewportUpdate)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setDragMode(QtWidgets.QGraphicsView.NoDrag)

    def mousePressEvent(self, event):
        if (event.button() == QtCore.Qt.LeftButton):
            item = self.itemAt(event.pos())
            if isinstance(item, NodeSocket):
                self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
            elif isinstance(item, Node):
                self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
            else:
                self.setDragMode(QtWidgets.QGraphicsView.RubberBandDrag)
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
        if self.scene():
            self.scene().node_selection_changed()
        super().mouseReleaseEvent(event)

    def wheelEvent(self, event):
        zoom_factor = 1.1
        if (event.angleDelta().y() > 0):
            self.scale(zoom_factor, zoom_factor)
        else:
            self.scale((1 / zoom_factor), (1 / zoom_factor))

class NodeConnection(QtWidgets.QGraphicsPathItem):

    def __init__(self, start_socket, end_socket=None):
        super().__init__()
        self.start_socket = start_socket
        self.end_socket = end_socket
        self.start_pos = None
        self.end_pos = None
        if self.start_socket:
            self.start_socket.connections.append(self)
        if self.end_socket:
            self.end_socket.connections.append(self)
        self._pen = QtGui.QPen(QtGui.QColor(ColorPalette.instance().CONNECTION_LINE))
        self._pen.setWidth(2)
        self.setPen(self._pen)
        self.updatePath()

    def updatePath(self):
        'Update the path of the connection'
        if (not self.start_socket):
            return
        self.start_pos = self.start_socket.getSubSocketPos(self)
        self.end_pos = (self.end_socket.getSubSocketPos(self) if isinstance(self.end_socket, NodeSocket) else self.end_socket)
        path = QtGui.QPainterPath(self.start_pos)
        if (self.start_pos and self.end_pos):
            distance = (self.end_pos.x() - self.start_pos.x())
            ctrl1 = QtCore.QPointF((self.start_pos.x() + (distance * 0.5)), self.start_pos.y())
            ctrl2 = QtCore.QPointF((self.end_pos.x() - (distance * 0.5)), self.end_pos.y())
            path.cubicTo(ctrl1, ctrl2, self.end_pos)
        self.setPath(path)

    def finalize(self):
        'Called when a connection is completed to add it to both sockets'
        if isinstance(self.end_socket, NodeSocket):
            if self.end_socket.addConnection(self):
                self.updatePath()
                return True
        return False

    def remove(self):
        log.debug(f'Removing connection {self}')
        try:
            if self.start_socket:
                try:
                    self.start_socket.removeConnection(self)
                except Exception as e:
                    log.warning(f'Error removing connection from start socket: {e}')
            if isinstance(self.end_socket, NodeSocket):
                try:
                    self.end_socket.removeConnection(self)
                except Exception as e:
                    log.warning(f'Error removing connection from end socket: {e}')
            if self.scene():
                self.scene().removeItem(self)
        except Exception as e:
            log.warning(f'Error removing connection: {e}')

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu()
        remove_action = menu.addAction('Remove Connection')
        remove_action.triggered.connect(self.remove)
        menu.exec_(event.screenPos())

    def serialize(self):
        'Serialize connection data'
        if (not isinstance(self.end_socket, NodeSocket)):
            return
        data = {'start_node': self.start_socket.parentItem().id, 'start_socket': self.start_socket.serialize()['index'], 'start_socket_type': ('output' if self.start_socket.is_output else 'input'), 'end_node': self.end_socket.parentItem().id, 'end_socket': self.end_socket.serialize()['index'], 'end_socket_type': ('output' if self.end_socket.is_output else 'input')}
        return data

class NodeScene(QtWidgets.QGraphicsScene):

    def __init__(self):
        super().__init__()
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(ColorPalette.instance().SCENE_BACKGROUND)))
        self.current_connection = None
        self.connecting_socket = None
        self.temp_connection = None
        self.on_selection_changed = None

    def serialize(self):
        'Serialize the entire scene to JSON'
        nodes = [item for item in self.items() if isinstance(item, Node)]
        connections = [item for item in self.items() if isinstance(item, NodeConnection)]
        nodes_data = [node.serialize() for node in nodes]
        array_connections = []
        other_connections = []
        for conn in connections:
            if (isinstance(conn.end_socket, NodeSocket) and conn.end_socket.is_array and (conn in conn.end_socket._connection_to_socket)):
                array_connections.append(conn)
            else:
                other_connections.append(conn)
        array_connections.sort(key=(lambda c: c.end_socket._connection_to_socket[c]))
        connections_data = []
        for conn in (array_connections + other_connections):
            conn_data = conn.serialize()
            if conn_data:
                connections_data.append(conn_data)
        return {'nodes': nodes_data, 'connections': connections_data}

    def deserialize(self, data):
        'Restore the scene from serialized data'
        self.clear()
        nodes = {}
        for node_data in data['nodes']:
            node_class = Node.registry.get(node_data['type'])
            if (not node_class):
                print(f"Warning: Node type {node_data['type']} not found in registry")
                continue
            node = node_class.deserialize(node_data, self)
            nodes[node_data['id']] = node
        for conn_data in data['connections']:
            start_node = nodes.get(conn_data['start_node'])
            end_node = nodes.get(conn_data['end_node'])
            if (not (start_node and end_node)):
                continue
            start_socket = (start_node.output_sockets if (conn_data['start_socket_type'] == 'output') else start_node.input_sockets)[conn_data['start_socket']]
            end_socket = (end_node.output_sockets if (conn_data['end_socket_type'] == 'output') else end_node.input_sockets)[conn_data['end_socket']]
            connection = NodeConnection(start_socket, end_socket)
            self.addItem(connection)
            connection.finalize()
            connection.updatePath()

    def startConnection(self, socket):
        'Start creating a new connection from the given socket'
        self.connecting_socket = socket
        self.temp_connection = NodeConnection(socket)
        self.addItem(self.temp_connection)

    def node_selection_changed(self):
        'Called when node selection changes'
        if self.on_selection_changed:
            selected_nodes = [item for item in self.selectedItems() if isinstance(item, Node)]
            self.on_selection_changed(selected_nodes)

    def mousePressEvent(self, event):
        clicked_item = self.itemAt(event.scenePos(), QtGui.QTransform())
        if (event.button() == QtCore.Qt.LeftButton):
            if (not clicked_item):
                self.clearSelection()
                self.node_selection_changed()
            elif isinstance(clicked_item, NodeSocket):
                self.startConnection(clicked_item)
        elif (event.button() == QtCore.Qt.RightButton):
            event.accept()
            return
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self.temp_connection:
            self.temp_connection.end_socket = event.scenePos()
            self.temp_connection.updatePath()
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self.connecting_socket:
            items = self.items(event.scenePos())
            end_item = None
            for item in items:
                if isinstance(item, NodeSocket):
                    end_item = item
                    break
            if (end_item and self.connecting_socket.can_connect_to(end_item)):
                connection = self.temp_connection
                connection.end_socket = end_item
                if connection.finalize():
                    self.temp_connection = None
                else:
                    self.removeItem(connection)
            else:
                self.removeItem(self.temp_connection)
            self.connecting_socket = None
            self.temp_connection = None
        super().mouseReleaseEvent(event)

    def clear(self):
        'Override clear to safely remove all items'
        for item in self.items():
            if isinstance(item, NodeConnection):
                try:
                    if item.start_socket:
                        item.start_socket.connections.remove(item)
                    if isinstance(item.end_socket, NodeSocket):
                        item.end_socket.connections.remove(item)
                except Exception as e:
                    log.warning(f'Error clearing connection references: {e}')
        for item in self.items():
            if isinstance(item, Node):
                try:
                    item.dcc_object = None
                except Exception as e:
                    log.warning(f'Error clearing node DCC reference: {e}')
        super().clear()
