from base64 import b64decode as license64

import os
from maya import cmds
from BroTools.core.broqt.vendor.Qt import QtCore, QtGui, QtWidgets
from BroTools.configuration import modpath
from BroTools.core.debug import default_logger as log
from BroTools.core import maya_utils
from BroTools.core.system.utilities import BroNode, convertListToBroNodes, isBroNodeInstance
rp = os.path.dirname(os.path.realpath(__file__))

class ObjectListWidget(QtWidgets.QWidget):
    'Custom widget for managing Maya object lists'

    def __init__(A, parent, title, object_type_filter=None, icon_path=None, tooltip_filter=None):
        "\n        Initialize an object list widget\n        \n        Args:\n            parent: Parent widget\n            title: Title of the list\n            object_type_filter: Type(s) of objects to accept:\n                               - None for any objects\n                               - A string for single object type (e.g., 'mesh')\n                               - A list of strings for multiple types (e.g., ['mesh', 'nurbsCurve'])\n            icon_path: Path to the icon for list items\n            tooltip_filter: Event filter for tooltips\n        "
        B = parent
        super(ObjectListWidget, A).__init__(B)
        A.parent = B
        A.title = title
        A.object_type_filter = object_type_filter
        A.icon_path = icon_path
        A.tooltip_filter = tooltip_filter
        A.buildUI()

    def buildUI(A):
        'Build the UI for this widget'
        A.layout = QtWidgets.QVBoxLayout(A)
        A.setLayout(A.layout)
        A.listLabel = QtWidgets.QLabel(f'{A.title}:')
        A.layout.addWidget(A.listLabel)
        A.objectsList = QtWidgets.QListWidget()
        A.objectsList.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        A.objectsList.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        A.objectsList.customContextMenuRequested.connect(A.showListContextMenu)
        A.layout.addWidget(A.objectsList)
        A.buttonsLayout = QtWidgets.QHBoxLayout()
        A.addObjectBtn = QtWidgets.QPushButton(f'Add {A.title}')
        A.addObjectBtn.clicked.connect(A.addSelectedObjects)
        if A.tooltip_filter:
            A.addObjectBtn.installEventFilter(A.tooltip_filter)
            maya_utils.setBroToolTipData(A.addObjectBtn, 'text', f'Select object(s) you want to add as {A.title.lower()} and click this button.')
        A.buttonsLayout.addWidget(A.addObjectBtn)
        A.removeBtn = QtWidgets.QPushButton('Remove')
        B = os.path.join(modpath, 'imgs', 'i16_remove.png')
        if os.path.exists(B):
            A.removeBtn.setIcon(QtGui.QIcon(B))
        A.removeBtn.clicked.connect(A.removeSelectedObjects)
        if A.tooltip_filter:
            A.removeBtn.installEventFilter(A.tooltip_filter)
            maya_utils.setBroToolTipData(A.removeBtn, 'text', f'Remove selected {A.title.lower()} from the list.')
        A.buttonsLayout.addWidget(A.removeBtn)
        A.clearBtn = QtWidgets.QPushButton('Clear All')
        C = os.path.join(modpath, 'imgs', 'i16_clear.png')
        if os.path.exists(C):
            A.clearBtn.setIcon(QtGui.QIcon(C))
        A.clearBtn.clicked.connect(A.clearList)
        if A.tooltip_filter:
            A.clearBtn.installEventFilter(A.tooltip_filter)
            maya_utils.setBroToolTipData(A.clearBtn, 'text', f'Clear all {A.title.lower()} from the list.')
        A.buttonsLayout.addWidget(A.clearBtn)
        A.layout.addLayout(A.buttonsLayout)

    def addSelectedObjects(A):
        'Add selected objects to the list'
        M = convertListToBroNodes(cmds.ls(sl=True))
        for E in M:
            J = E.name
            K = E.long_name
            B = True
            if (A.object_type_filter is not None):
                B = False
                D = cmds.objectType(E.long_name)
                if isinstance(A.object_type_filter, str):
                    if (A.object_type_filter == 'mesh'):
                        if (D == 'transform'):
                            C = cmds.listRelatives(E.long_name, shapes=True, fullPath=True)
                            if C:
                                for F in C:
                                    if (cmds.objectType(F) == 'mesh'):
                                        B = True
                                        break
                        elif (D == 'mesh'):
                            B = True
                    else:
                        B = (D == A.object_type_filter)
                elif isinstance(A.object_type_filter, list):
                    if ('mesh' in A.object_type_filter):
                        if (D == 'transform'):
                            C = cmds.listRelatives(E.long_name, shapes=True, fullPath=True)
                            if C:
                                for F in C:
                                    I = cmds.objectType(F)
                                    if (I in A.object_type_filter):
                                        B = True
                                        break
                        elif (D in A.object_type_filter):
                            B = True
                    else:
                        B = (D in A.object_type_filter)
                        if ((not B) and (D == 'transform')):
                            C = cmds.listRelatives(E.long_name, shapes=True, fullPath=True)
                            if C:
                                for F in C:
                                    I = cmds.objectType(F)
                                    if (I in A.object_type_filter):
                                        B = True
                                        break
                if (not B):
                    G = A.object_type_filter
                    if isinstance(G, list):
                        G = ', '.join(G)
                    log.warning(f'Object {J} does not match type filter ({G}) and cannot be used as a {A.title.lower()}.')
            if B:
                H = QtWidgets.QListWidgetItem()
                H.setText(J)
                H.setData(QtCore.Qt.UserRole, K)
                if (A.icon_path and os.path.exists(A.icon_path)):
                    H.setIcon(QtGui.QIcon(A.icon_path))
                L = False
                for N in range(A.objectsList.count()):
                    if (A.objectsList.item(N).data(QtCore.Qt.UserRole) == K):
                        L = True
                        break
                if (not L):
                    A.objectsList.addItem(H)

    def showListContextMenu(A, position):
        'Show context menu for the list'
        B = QtWidgets.QMenu()
        C = B.addAction('Remove')
        C.setIcon(QtGui.QIcon(os.path.join(modpath, 'imgs', 'i16_remove.png')))
        C.triggered.connect(A.removeSelectedObjects)
        D = B.addAction('Select in Maya')
        D.triggered.connect(A.selectObjectInMaya)
        B.exec_(A.objectsList.mapToGlobal(position))

    def removeSelectedObjects(A):
        'Remove selected objects from the list'
        for B in A.objectsList.selectedItems():
            A.objectsList.takeItem(A.objectsList.row(B))

    def clearList(A):
        'Clear all objects from the list'
        A.objectsList.clear()

    def selectObjectInMaya(B):
        'Select the object in Maya scene'
        A = []
        for C in B.objectsList.selectedItems():
            D = C.data(QtCore.Qt.UserRole)
            A.append(D)
        if A:
            cmds.select(A, replace=True)

    def getObjects(A):
        'Get all objects in the list'
        B = []
        for C in range(A.objectsList.count()):
            D = A.objectsList.item(C)
            E = D.data(QtCore.Qt.UserRole)
            B.append(E)
        return B
