from base64 import b64decode as license64

'\nUtils that interact with Maya and need the use of BroNode.\n'
import base64, copy, math, os, ntpath, sys, time, traceback, uuid, webbrowser, zlib, re, ast, _ast
from functools import partial
from pprint import PrettyPrinter
pp = PrettyPrinter(indent=4)
from maya.api import OpenMaya as om1
import maya.OpenMaya as om, maya.api.OpenMayaAnim as oma
from maya import cmds
from maya import mel
from BroTools.configuration import modpath, icon, APP_DIR, MOD_NAME, PY3, modulepath_to_filepath, filepath_to_modulepath, brotools_config
from BroTools.core.debug import default_logger as log
from BroTools.core.errors import BroToolsError, catch_errors
from BroTools.core.system.utilities import BroNode, ensureBroNode, smart_pyopen, Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07, convertListToBroNodes
from BroTools.core.utils import *
from BroTools.lib.euclid import Vector3 as Vector3D, Matrix4, Quaternion
if PY3:
    import importlib
    reload = importlib.reload
    import builtins
    builtins.unicode = str
BRO_springProxySuffix = '_SimProxyBro'
refresh = False
refresh_delay = 0.02

def do_refresh():
    if refresh:
        cmds.refresh()
        time.sleep(refresh_delay)

def add_namespace_to_longName(longName, namespace):
    log.debug('Add namespace "{}" to ""'.format(namespace, longName))
    if ((namespace == '') or (namespace == None)):
        return longName
    newLongName = ''
    objects = longName.split('|')
    log.debug('NS objects', objects)
    for (i, obj) in enumerate(objects):
        if (obj != ''):
            newLongName += '|{}:{}'.format(namespace, obj)
            log.debug('Adding namespace...', newLongName)
    log.debug('Namespace added:', newLongName)
    return newLongName

def remove_namespaces_from_longName(longName):
    newLongName = ''
    objects = longName.split('|')
    for (i, obj) in enumerate(objects):
        if (i > 0):
            newLongName += obj.rpartition(':')[(- 1)]
    return newLongName

def list_namespaces():
    current_namespace = cmds.namespaceInfo(cur=True)
    cmds.namespace(setNamespace=':')
    namespaces = cmds.namespaceInfo(listOnlyNamespaces=True, recurse=True)
    cmds.namespace(setNamespace=current_namespace)
    return namespaces

def clamp(n, smallest, largest):
    return max(smallest, min(n, largest))

def mvector_to_tuple(vec):
    return (vec.x, vec.y, vec.z)

@catch_errors
def matchTransform(child, parent, t=True, r=True, offset=[], **kwargs):
    if isinstance(child, list):
        if (len(child) > 1):
            raise BroToolsError('matchTransform child object must be a string or BroNode, list found: {}'.format(child))
        else:
            child = BroNode(child)
    lock_filter = kwargs.get('lock_filter', True)
    try:
        if lock_filter:
            for ax in 'xyz':
                if cmds.getAttr(((child + '.t') + ax), lock=True):
                    t = False
                if cmds.getAttr(((child + '.r') + ax), lock=True):
                    r = False
        cmds.matchTransform(child, parent, pos=t, rot=r, piv=(True if ((brotools_config.MayaUtils.matchTransform_MatchPivot is True) and (t is True)) else False))
        return
    except Exception as e:
        reason = e
        traceback.print_exc()
    try:
        log.info('cmds.matchTransform not supported, using old openMaya and cmds.xform based solution. Reason:', reason)
        source_selList = om1.MSelectionList()
        source_selList.add(parent)
        source_mDagPath = om1.MDagPath()
        source_selList.getDagPath(0, source_mDagPath)
        source = om1.MFnTransform(source_mDagPath)
        destin_selList = om1.MSelectionList()
        destin_selList.add(child)
        destin_mDagPath = om1.MDagPath()
        destin_selList.getDagPath(0, destin_mDagPath)
        destin = om1.MFnTransform(destin_mDagPath)
        if t:
            rotPivA = om1.MVector(source.rotatePivot(om1.MSpace.kWorld))
            rotPivB = om1.MVector(destin.rotatePivot(om1.MSpace.kWorld))
            origTrans = destin.translation(om1.MSpace.kWorld)
            targetTrans = (rotPivA + (origTrans - rotPivB))
            destin.setTranslation(targetTrans, om1.MSpace.kWorld)
        if r:
            origRots = om1.MQuaternion()
            destin.rotation(om1.MSpace.kWorld, origRots)
            Quat = om1.MQuaternion()
            source.rotation(om1.MSpace.kWorld, Quat)
            destin.setRotation(Quat, om1.MSpace.kWorld)
    except:
        if t:
            cmds.xform(child, ws=True, t=cmds.xform(parent, q=True, ws=True, t=True))
        if r:
            cmds.xform(child, ws=True, ro=cmds.xform(parent, q=True, ws=True, ro=True))

@catch_errors
def scaleSpaceLocator(locator, scale):
    for axis in ['X', 'Y', 'Z']:
        cmds.setAttr(((locator + '.localScale') + axis), scale)

@catch_errors
def getTimelineSelectedRange(*args, **kwargs):
    aTimeSlider = mel.eval('$tmpVar=$gPlayBackSlider')
    timeRange = cmds.timeControl(aTimeSlider, q=True, rangeArray=True)
    return (int(timeRange[0]), int(timeRange[1]))

@catch_errors
def getCurrentTimeRange(respect_selection=True, single_frame_ranges=False):
    '\n\n    :param respect_selection: If True will respect current timeline selection (with Shift+Drag)\n    :param single_frame_ranges: If False will select the whole timeline range instead of single selected frame\n    :return: (start, end) integers\n    '
    if respect_selection:
        selectedRange = getTimelineSelectedRange()
        if single_frame_ranges:
            return selectedRange
        if ((selectedRange[1] - selectedRange[0]) > 1):
            return selectedRange
    return (int(cmds.playbackOptions(q=True, min=True)), int(cmds.playbackOptions(q=True, max=True)))

@catch_errors
def getCurrentTimeRangeFrameCount(respect_selection=True, single_frame_ranges=False):
    (startFrame, endFrame) = getCurrentTimeRange(respect_selection, single_frame_ranges)
    frameCount = ((endFrame - startFrame) + 1)
    return frameCount

@catch_errors
def trackTransform(child=None, parent=None, t=True, r=True, timerange=None, key=True, progress=True, mo=False, match_objects_keyframes=None, euler_filter=False, fast=False, super_fast=True, children_parents=None, iterations=1, backwards=False, *args):
    if match_objects_keyframes:
        match_objects_keyframes = ensureList(match_objects_keyframes)
        log.debug('match_objects_keyframes: {}'.format(match_objects_keyframes))
        match_keyframes = getKeyframeTimes(match_objects_keyframes, integers=False)
    else:
        if (timerange == None):
            timerange = getCurrentTimeRange()
        match_keyframes = list(range(timerange[0], timerange[1]))
    if backwards:
        match_keyframes = list(reversed(match_keyframes))
    cmds.currentTime(match_keyframes[0])
    if (children_parents is None):
        children_parents = []
    if (isinstance(child, list) and isinstance(parent, list)):
        for (i, c) in enumerate(child):
            children_parents.append([c, parent[i]])
    elif (len(children_parents) == 0):
        children_parents.append(list(reversed(getParentChild(child, parent))))
    children = []
    for (child, parent) in children_parents:
        children.append(child)
    log.debug('Track transform:', children_parents, t, r, timerange, key, progress, mo, match_objects_keyframes, euler_filter, fast)
    if (fast and (not match_objects_keyframes)):
        log.debug('trackTransform fast mode')
        for (child, parent) in children_parents:
            if (t and r):
                cmds.parentConstraint(parent, child, mo=mo)
            else:
                if t:
                    cmds.pointConstraint(parent, child, mo=mo)
                if r:
                    cmds.orientConstraint(parent, child, mo=mo)
        cmds.bakeResults(children, t=getCurrentTimeRange(), sm=(not super_fast))
        for child in children:
            cmds.delete(child, constraints=True)
        return
    if mo:
        for (i, cp) in enumerate(children_parents):
            (child, parent) = cp
            loc = cmds.spaceLocator()[0]
            log.debug('Using offset locator:', loc)
            matchTransform(loc, child, t=True, r=True)
            cmds.parentConstraint(parent, loc, mo=True)
            children_parents[i] = [child, copy.copy(loc)]
    log.debug('match_keyframes: {}'.format(match_keyframes))
    for frame in match_keyframes:
        cmds.currentTime(frame)
        for iteration in range(iterations):
            for (child, parent) in children_parents:
                matchTransform(child, parent, t, r)
                if key:
                    for a in ['t', 'r']:
                        for x in ['x', 'y', 'z']:
                            if (not cmds.getAttr((((child + '.') + a) + x), lock=True)):
                                cmds.setKeyframe(child, attribute=(a + x))
    if euler_filter:
        cmds.filterCurve(child, f='euler')
    if mo:
        log.debug('Delete locator', loc)
        cmds.delete(loc)
'\ndef transferMotion(child=None, parent=None, mo=True):\n    parent, child = getParentChild(child, parent)\n\n    dummy = BroNode.create(\'spaceLocator\')\n    log.info("DUMMY {}".format(dummy))\n    matchTransform(dummy, child)\n    trackTransform(dummy, parent, mo=mo, match_objects_keyframes=parent)\n    trackTransform(child, dummy, mo=False, match_objects_keyframes=dummy)\n\n    dummy.delete()\n'

def getParentChild(child=None, parent=None, inverse=False, bronode=True):
    '\n    If no child or parent specified tries to get them from selection, and if none available returns a maya error.\n\n    inverse: Reverse order of parent child\\child parent\n    bronode: Use False to not cast to bronode class, may be useful for performance optimization\n    '
    selection = cmds.ls(sl=True)
    if ((child == None) and (parent == None) and (len(selection) == 0)):
        cmds.error('Please, select parent and child before using trackTransform command.')
    if (parent == None):
        parent = selection[(1 if inverse else 0)]
    if (child == None):
        child = selection[(0 if inverse else 1)]
    if bronode:
        return (BroNode(parent), BroNode(child))
    else:
        return (parent, child)

def aimTransform(child=None, target=None, aimAxis='x', upAxis='y', worldUpObject=None):
    '\n    Deprecated old attempts at aiming through math only. Superceeded by aim_object.\n    Over time this should be removed and replaced with aim_object.\n    Never worked properly.\n    '
    aimAxis = str(aimAxis).lower()
    upAxis = str(upAxis).lower()
    if (aimAxis not in ['x', 'y', 'z']):
        raise BroToolsError(('aimTransform error, invalid aimAxis:' + str(aimAxis)))
    if (upAxis not in ['x', 'y', 'z']):
        raise BroToolsError(('aimTransform error, invalid upAxis:' + str(upAxis)))
    childAim = om1.MVector().__getattribute__((aimAxis + 'Axis'))
    childUp = om1.MVector().__getattribute__((upAxis + 'Axis'))
    child = BroNode(child)
    target = BroNode(target)
    transformFn = om1.MFnTransform(child._mDagPath)
    childPivotPOS = transformFn.rotatePivot(om1.MSpace.kWorld)
    transformFn.setObject(target._mDagPath)
    targetPivotPOS = transformFn.rotatePivot(om1.MSpace.kWorld)
    aimVector = (targetPivotPOS - childPivotPOS)
    childU = aimVector.normal()
    if (worldUpObject is not None):
        worldUpObject = BroNode(worldUpObject)
        transformFn.setObject(worldUpObject._mDagPath)
        worldUpObjectPivotPOS = transformFn.rotatePivot(om1.MSpace.kWorld)
        worldUp = (childPivotPOS - worldUpObjectPivotPOS)
    else:
        worldUp = om1.MGlobal.upAxis()
    childV = worldUp
    childW = (childU ^ childV).normal()
    childV = (childW ^ childU)
    quaternionU = om1.MQuaternion(childAim, childU)
    quaternion = quaternionU
    upRotate = childUp.rotateBy(quaternion)
    angle = math.acos(clamp((upRotate * childV), (- 1), 1))
    quaternionV = om1.MQuaternion(angle, childU)
    if (not childV.isEquivalent(upRotate.rotateBy(quaternionV), 1e-05)):
        angle = ((2 * math.pi) - angle)
        quaterntionV = om1.MQuaternion(angle, childU)
    quaternion *= quaternionV
    transformFn.setObject(child._mDagPath)
    transformFn.setRotation(quaternion)

def aim_object(source, target, forward_axis=Vector3D(1, 0, 0), up_axis=Vector3D(0, 1, 0), world_up=Vector3D(0, 1, 0), world_up_object=None):
    '\n    Aim source object at target object using pyEuclid and matrices.\n\n    This has been tested and it can be assume that this function works correctly, as well as Quaternion.look_at.\n    Axes are properly defined as well.\n    '
    if (world_up_object is not None):
        up_matrix = Matrix4.new(*cmds.xform(world_up_object, q=True, matrix=True, ws=True))
        world_up = up_matrix.get_quaternion().normal(up_axis)
    source_matrix = Matrix4.new(*cmds.xform(source, q=True, matrix=True, ws=True))
    source_pos = Vector3D(*cmds.xform(source, q=True, t=True, ws=True))
    target_pos = Vector3D(*cmds.xform(target, q=True, t=True, ws=True))
    direction = (target_pos - source_pos).normalize()
    quat = Quaternion.look_at(direction, forward_axis, up_axis, world_up)
    rot_matrix = quat.get_matrix()
    final_matrix = Matrix4()
    final_matrix.a = rot_matrix.a
    final_matrix.b = rot_matrix.b
    final_matrix.c = rot_matrix.c
    final_matrix.e = rot_matrix.e
    final_matrix.f = rot_matrix.f
    final_matrix.g = rot_matrix.g
    final_matrix.i = rot_matrix.i
    final_matrix.j = rot_matrix.j
    final_matrix.k = rot_matrix.k
    final_matrix.d = source_matrix.d
    final_matrix.h = source_matrix.h
    final_matrix.l = source_matrix.l
    cmds.xform(source, matrix=list(final_matrix), ws=True)

def positionOrObject(inObj):
    '\n    Checks if given object is a position tuple, if not - converts it to position tuple.\n    '
    if ((not isinstance(inObj, list)) and (not isinstance(inObj, tuple))):
        obj = BroNode(inObj).getTransform()
        obj = om1.MVector(obj.rotatePivot(om1.MSpace.kWorld))
        obj = (obj.x, obj.y, obj.z)
    else:
        obj = inObj
    if (len(obj) == 3):
        log.debug(inObj, 'converted to', obj)
        return obj
    else:
        log.error("Can't convert", obj, 'to position tuple.')
        return

def placeObjectInTheMiddle(obj, verts):
    'Works for objects or verts'
    vtxPosList = []
    vectPos = []
    sumVect = om1.MVector(0, 0, 0)
    if (verts == None):
        selList = cmds.ls(sl=True)
    else:
        selList = verts
    for x in selList:
        vtxPosList.extend(cmds.xform(x, q=True, ws=True, translation=True))
    for x in range(0, int((len(vtxPosList) / 3))):
        vect = om1.MVector(vtxPosList[(x * 3)], vtxPosList[((x * 3) + 1)], vtxPosList[((x * 3) + 2)])
        vectPos.append(vect)
        sumVect += vect
    sumVect = (sumVect / (len(vtxPosList) / 3))
    cmds.xform(obj, ws=True, translation=[sumVect.x, sumVect.y, sumVect.z])
moveObjectInTheMiddle = placeObjectInTheMiddle

@catch_errors
def extendByVector(objA, objB, target=None):
    "\n    Find extending position based on vectors of two objects.\n    Extending position is B + (B - A)\n    Moved target object extending it's position based on objA and objB vector.\n\n    :return x, y, z tuple, location in world space\n    "
    A = om1.MVector(*positionOrObject(objA))
    B = om1.MVector(*positionOrObject(objB))
    new_pos = (B + (B - A))
    log.debug(objA, objB, [A.x, A.y, A.z], [B.x, B.y, B.z], [new_pos.x, new_pos.y, new_pos.z])
    if (target != None):
        tg = BroNode(target).getTransform()
        tg.setTranslation(new_pos, om1.MSpace.kWorld)
    return (new_pos.x, new_pos.y, new_pos.z)

@catch_errors
def extendByAxis(objA, objB, target=None, axis=[1, 0, 0], multiplier=1, distance=None):
    "\n    Find extending position based on extending along existing object's axis.\n\n    :return x, y, z tuple, location in world space\n    "
    if (distance == None):
        distance = getDistance(objA, objB)
    if (axis == 'auto'):
        axis = axis_letter_to_array(axisFacingObject(objA, objB))
    distance = (distance * multiplier)
    translation = []
    for ax in axis:
        translation.append((ax * distance))
    log.debug('Translate', translation)
    loc = cmds.spaceLocator()
    matchTransform(loc, objA, t=True, r=True)
    cmds.xform(loc, relative=True, objectSpace=True, t=translation)
    new_pos = getWSPosition(loc)
    if (target != None):
        matchTransform(target, loc, t=True, r=True)
    cmds.delete(loc)
    return new_pos

def getLocalVecToWorldSpace(node, vec=om1.MVector.kXaxisVector):
    matrix = BroNode(node)._mDagPath.inclusiveMatrix()
    vec = (vec * matrix).normal()
    return vec

def axisVectorColinearity(node, vec):
    '\n    Find which axis is facing world space position the most. Returns dict with values.\n    TODO: Check with older maya versions, xAxis, y,z may be named differently there\n    '
    vec = om1.MVector(*vec)
    x = getLocalVecToWorldSpace(node, vec=om1.MVector.kXaxisVector)
    y = getLocalVecToWorldSpace(node, vec=om1.MVector.kYaxisVector)
    z = getLocalVecToWorldSpace(node, vec=om1.MVector.kZaxisVector)
    return {'x': (vec * x), 'y': (vec * y), 'z': (vec * z)}

def _axisFacingObject(node, vec):
    'Returns which axis is facing object.'
    data = axisVectorColinearity(node, vec)
    return max(data, key=data.get)

def axisFacingObject(node, target):
    'Returns which axis is facing object.'
    vec = cmds.xform(target, q=True, rp=True, ws=True)
    data = axisVectorColinearity(node, vec)
    return max(data, key=data.get)

def snapMeshObjectToGroundByClosestPolygonFace(objects=None):
    if (objects is None):
        objects = cmds.ls(selection=True)
    for object in objects:
        faces = (cmds.ls(object, type='mesh', dag=True, shapes=True)[0] + '.f[*]')
        centers = cmds.xform(faces, q=True, t=True, ws=True)
        min_y = float('inf')
        for i in range(0, len(centers), 3):
            y = centers[(i + 1)]
            if (y < min_y):
                min_y = y
        distance = abs(min_y)
        if (min_y > 0):
            cmds.move(0, (- distance), 0, object, relative=True)
        else:
            cmds.move(0, distance, 0, object, relative=True)

def findSkinClusters(meshes=None):
    if (meshes is None):
        meshes = cmds.ls(sl=1)
    if (not isinstance(meshes, list)):
        meshes = [meshes]
    skin_clusters = []
    for mesh in meshes:
        shapes = cmds.listRelatives(mesh, c=True, s=True, f=True)
        for shape in shapes:
            sk = mel.eval(('findRelatedSkinCluster ' + shape))
            if (not sk.strip()):
                sk = None
            log.debug('// Mesh: {}; Shape: {}; skinCluster: {};'.format(mesh, shape, sk))
            if (sk is not None):
                skin_clusters.append(sk)
    return skin_clusters

def findSkinInfluencingJoints(meshes=None):
    joints = []
    skin_clusters = findSkinClusters(meshes)
    for sk in skin_clusters:
        influences = cmds.skinCluster(sk, q=True, inf=True)
        joints.extend(influences)
        log.debug('// Found more influencing joints on ({}): {}'.format(sk, influences))
    return joints

def selectSkinInfluencingJoints(meshes=None):
    joints = findSkinInfluencingJoints(meshes)
    cmds.select(joints, r=True)
    log.debug('Influencing joints:')
    log.debug(joints)
    log.debug('---')

def parse_attr(a):
    '\n    Ensures that passed attribute is in maya\'s format "name.attr".\n    Can take in pairs [BroNode, \'attr\'] or string \'name.attr\', or pair [\'name\', \'attr\']\n\n    TODO Make it possible to pass BroNode.attr with no quotes.\n    '
    if isinstance(a, list):
        if isinstance(a[0], BroNode):
            a_name = a[0].longName
        else:
            a_name = a[0]
        a = ((a_name + '.') + a[1])
    return a

def connect(*args, **kwargs):
    '\n    Connects the first attribute to the second.\n    TODO Improve to remove rudimentary parse_attr, it only adds overhead, or make it work with bronodes properly\n    '
    log.debug('Connect: {}; {};'.format(str(args), str(kwargs)))
    if (len(args) == 2):
        a = args[0]
        b = args[1]
        if (isinstance(a, list) and isinstance(b, list)):
            log.debug('Connecting {} and {}: {} {}'.format(a, b, parse_attr(a), parse_attr(b)))
            result = cmds.connectAttr(parse_attr(a), parse_attr(b), **kwargs)
            log.debug(result)
            return result
    elif (len(args) == 4):
        a = [args[0], args[1]]
        b = [args[2], args[3]]
        log.debug('Connecting {} and {}: {} {}'.format(a, b, parse_attr(a), parse_attr(b)))
        result = cmds.connectAttr(parse_attr(a), parse_attr(b), **kwargs)
        log.debug(result)
        return result
    cmds.error('Cant connect nodes, wrong number of parameters?')
connectAttr = connect_attr = connect_attribute = connect

def connectNodes(*args, **kwargs):
    "\n    connectNodes(\n        {\n            a:\n                [b, 'sdsd', 'sdsd'],\n                [c, 'sdsd', 'sdsd']\n        }\n    )\n    "
    log.debug('connectNodes: {args}; {details};'.format(args=str(args), details=str(kwargs)))
    con_dict = args[0]
    for (node, connections) in con_dict.items():
        for connection in connections:
            node = ensureBroNode(node)
            node.connect(*connection)

def disconnect(a, b):
    '\n    Dionnects the first atribute from the second.\n    '
    return cmds.disconnectAttr(parse_attr(a), parse_attr(b))

def lock(a):
    return cmds.setAttr(parse_attr(a), lock=True)

def unlock(a):
    return cmds.setAttr(parse_attr(a), lock=False)

def maAddAttrValueType(value):
    '\n    This function returns a value type suitable for maya setAttr\x07ddAttr\\getAttr commands.\n    Returns dict that can be **passed to a function.\n    :param value: Value\n    :return:\n    '
    type_values = {str: {'dt': 'string'}, bool: {'at': 'boolean'}, int: {'at': 'long'}, float: {'at': 'float'}}
    value_type = type(value)
    if (value_type in type_values):
        return type_values[value_type]
    else:
        return {}

def maSetAttrValueType(value):
    '\n    This function returns a value type suitable for maya setAttr\x07ddAttr\\getAttr commands.\n    Returns dict that can be **passed to a function.\n    :param value: Value\n    :return:\n    '
    type_values = {str: {'type': 'string'}}
    value_type = type(value)
    if (value_type in type_values):
        return type_values[value_type]
    else:
        return {}

@catch_errors
def findByAttr(attr, value, type=None, mult=False):
    '\n    Find object by attribute.\n    Full search for the whole scene with 10000 objects takes about 0.5 seconds.\n    Limiting it to one object type takes about 0.2 seconds (discarding secondary nodes for each object, and utility nodes)\n    Finding only single object (not mult) can take anywhere between 0 and 0.2\x00.5 seconds.\n\n    :param attr: Attribute to check\n    :param value: Value which this attribute should have\n    :param type: (Optimization) Limit cmds.ls query to this type of objects. Speeds up search.\n    :param mult: If True - return first found object. If False - gather and return all matching objects. Finding just one object is usually faster.\n    :return: Returns single object name for single object search\n    Returns array of objects for mult search\n    Returns None if nothing is found.\n    '
    if (type == None):
        objects = cmds.ls()
    else:
        objects = cmds.ls(type=type)
    return_objs = []
    for obj in objects:
        obj_attr = ((obj + '.') + attr)
        if cmds.objExists(obj_attr):
            if (cmds.getAttr(obj_attr) == value):
                if (not mult):
                    return obj
                return_objs.append(obj)
    if (len(return_objs) == 0):
        return
    else:
        return return_objs

@catch_errors
def attrTypeFromValue(value):
    pass

def setAttrsFromDict(obj, settings):
    '\n    Deprecated, use BroNode instead.\n    Args:\n        RBDShape: Bullet RBD Shape\n        settings: DICT attribute:value\n    '
    for (attr, value) in settings.items():
        try:
            cmds.setAttr(((obj + '.') + attr), value)
        except Exception as e:
            log.info('!', 'Unable to set attr:', attr, value, ' Error:', e)
'\ndef do_parent(child, parent, ignore_errors=True):\n    if uid.isUUID(child):\n        child = uid.get_node(child)\n    if uid.isUUID(parent):\n        parent = uid.get_node(parent)\n\n    log.info(">< Parent", child, parent)\n    ret = None\n    prnt = cmds.listRelatives(child, p=True, f=True)\n    if prnt != None:\n        if prnt[0] == uid.get_node(parent):\n            if ignore_errors:\n                log.info("! Did not parent", child, parent, "It\'s already a child", lvl=3)\n                return ret\n            else:\n                cmds.error("! Did not parent", child, parent, "It\'s already a child")\n    ret = cmds.parent(child, parent)\n    do_refresh()\n\n    return ret\n\n\ndef do_constraint(p, c, parent=False, point=False, orient=False, scale=False, mo=False):\n    if uid.isUUID(c):\n        c = uid.get_node(c)\n    if uid.isUUID(p):\n        p = uid.get_node(p)\n\n    log.info("Constraining", c, "to", p, parent, point, orient, scale, mo)\n\n    constraints = {}\n    if parent:\n        constraints[\'parent\'] = uid.get_uuid(cmds.parentConstraint(p, c, mo=mo))\n    if point:\n        constraints[\'point\'] = uid.get_uuid(cmds.pointConstraint(p, c, mo=mo))\n    if orient:\n        constraints[\'orientation\'] = uid.get_uuid(cmds.orientConstraint(p, c, mo=mo))\n    if scale:\n        constraints[\'scale\'] = uid.get_uuid(cmds.scaleConstraint(p, c, mo=mo))\n    do_refresh()\n    return constraints\n\n\ndef do_refresh():\n    if refresh:\n        cmds.refresh()\n        time.sleep(refresh_delay)\n'
from BroTools.configuration import MOD_MenuRoot
import ast

def get_declaration_from_source(fp, name='__all__'):
    'gets a single declaration from python source code'
    with open(fp, 'r') as f:
        text = f.read()
    tree = ast.parse(text)
    for node in tree.body:
        if (isinstance(node, ast.Assign) and (len(node.targets) == 1)):
            target = node.targets[0]
            if (isinstance(target, ast.Name) and (target.id == name)):
                return ast.literal_eval(node.value)

def preprocess_menu_data(data):
    "\n    Process menu dict data from file to define default values.\n\n    Generates label from filename if 'label' is not passed in data dict.\n\n    :rtype: dict\n    "
    defaults = {'parent': MOD_MenuRoot}
    for (key, value) in data.items():
        if (value == 'default'):
            if (key in defaults):
                data[key] = defaults[key]
    if (data.get('label') is None):
        data['label'] = split_words_in_variable(data['command'].__name__).title()
    if ((data.get('ann') is None) and data['command'].__doc__):
        if isinstance(data['command'], partial):
            data['ann'] = data['command'].func.__doc__
        else:
            data['ann'] = data['command'].__doc__
    if ('icon' in data):
        data['image'] = icon(data['icon'], fp=True)
    else:
        data['image'] = icon(sluggify(data['label']))
    return data

def dummy_command():
    log.error('License invalid for this function.')

def add_menu_item(**kwargs):
    Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07 = kwargs.pop('Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07', None)
    data = preprocess_menu_data(kwargs)
    log.debug('add_menu_item', kwargs, data)
    if (data['label'] == None):
        log.info('No label specified for menu item.')
        return
    if cmds.menu(data['parent'], q=True, exists=True, aob=True):
        name = (data['label'].replace(' ', '_').lower() + '_menuitem')
        cmds.setParent(data['parent'], menu=True)
        if (Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07 is False):
            data['enable'] = False
            data['label'] += ' [Invalid License]'
            data['command'] = partial(dummy_command)
        cmds.menuItem(name, echoCommand=True, **data)
    else:
        log.info('Parent', data['parent'], "does not exist, can't create menu", data['label'])
import importlib

def get_variables(node):
    '\n    Use ast to get variables from a node, from a python file without importing it.\n    TODO: Move to utils\n    :param node: ast.parse(open(fp).read())\n    :return: set of vars\n    '
    variables = set()
    if hasattr(node, 'body'):
        for subnode in node.body:
            try:
                variables |= get_variables(subnode)
            except Exception as e:
                log.warning(e)
    elif isinstance(node, _ast.Assign):
        for name in node.targets:
            if isinstance(name, _ast.Name):
                variables.add(name.id)
    return variables

def get_current_time():
    return oma.MAnimControl.currentTime().asUnits(om.MTime.uiUnit())

def get_current_fps():
    return om.MTime(1, om.MTime.kSeconds).asUnits(om.MTime.uiUnit())

def frame_range(startFrame=None, endFrame=None, reset_physics=False, respect_selection=None, dry=False):
    '\n    Iterator that goes over frames\n    :param start:\n    :param framecount:\n    :param dry: Iterates over frames without changing currentTime in the scene\n    :return: frame\n    '
    if ((startFrame is None) and (endFrame is None)):
        if (respect_selection is None):
            (startFrame, endFrame) = getCurrentTimeRange()
        else:
            (startFrame, endFrame) = getCurrentTimeRange(respect_selection=respect_selection)
    frameCount = ((endFrame - startFrame) + 1)
    if reset_physics:
        cmds.currentTime((startFrame + 1))
        cmds.refresh()
        cmds.currentTime(startFrame)
        cmds.refresh()
    for frame in range(0, frameCount):
        if progressBarIsCanceled():
            log.warning('Interrupted by user.')
            break
        if (not dry):
            cmds.currentTime((startFrame + frame))
        (yield (frame, frameCount))

def menu_from_files(*args, **kwargs):
    '\n    It scans folder of a file this function was called from and generates menu based on MENU_DATA entries.\n    '
    m_path = None
    try:
        namespace = sys._getframe(1).f_globals
        path = normpath(os.path.dirname(namespace.get('__file__', None)))
        THIS_PATH = (((MOD_NAME + '.') + normpath(os.path.relpath(path, normpath(APP_DIR)).replace(os.path.sep, '.'))) + '.')
        log.debug('menu_from_files', path, THIS_PATH, APP_DIR, args, kwargs)
        modules = []
        for f in os.listdir(path):
            (name, ext) = os.path.splitext(f)
            if (ext.lower() == '.py'):
                modules.append((THIS_PATH + name))
        log.debug('  menu_from_files data:', modules)
        for m_path in modules:
            try:
                fp = modulepath_to_filepath(m_path)
                vars = get_variables(ast.parse(smart_pyopen(fp)))
                log.debug(('HAS_MENU' in vars), m_path, fp, vars)
                if (('HAS_MENU' in vars) or ('MENU_DATA' in vars)):
                    log.debug('    menu_from_files,', m_path, 'has MENU_DATA, try import and add')
                    m = importlib.import_module(m_path)
                    if isinstance(m.MENU_DATA, dict):
                        add_menu_item(Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07=Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07(namespace=namespace), **m.MENU_DATA)
                    elif isinstance(m.MENU_DATA, list):
                        for menu_data_item in m.MENU_DATA:
                            add_menu_item(Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07=Ox8r0_ZUp3ckVWUFhxX0NlWnFUMF9zWVV1VVZWQUNwR0JpSQ_ox07(namespace=namespace), **menu_data_item)
                    else:
                        log.warning("Unsupported MENU_DATA type in {}: {}, can't build menu".format(m_path, str(type(m.MENU_DATA))))
            except:
                log.warning(('    Unable to load menu for ' + str(m_path)), warn=True, traceback=True)
    except:
        log.warning(('Failed to load menu for ' + str(m_path)), warn=True, traceback=True)
import functools

def menu_item_wrap(f):

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if ((args[0] == False) and (args[1] == 'menu_trigger')):
            del args[0]
            del args[0]
        print('----------------')
        print(args, kwargs)
        return f(*args, **kwargs)
    return wrapped

def setBroToolTipData(obj, attr, value, setStatusTip=True, setWhatsThis=True):
    if (not hasattr(obj, 'broTooltip')):
        obj.broTooltip = {}
    obj.broTooltip[attr] = value
    if ((attr == 'text') and (type(value) == str)):
        if setStatusTip:
            obj.setStatusTip(value)
        if setWhatsThis:
            obj.setWhatsThis(value)

def sort_outliner_objects(objects=None):
    if (not objects):
        objects = cmds.ls(sl=True)
    if (not objects):
        return
    resorted = sorted(objects)
    for obj in resorted:
        try:
            cmds.reorder(obj, b=True)
        except Exception as e:
            log.warning("Can't move {}: {}".format(obj, str(e)))

def zgroup(objects, groupType='group', separate=False, parent=True):
    selection = cmds.ls(sl=True)
    cmds.select(cl=True)
    objects = convertListToBroNodes(objects)
    if separate:
        for obj in objects:
            zgrp = BroNode.create(groupType, name=numberedName('{}_zgrp'.format(obj.name)))
            matchTransform(zgrp, obj)
            if parent:
                obj.parent = zgrp
    else:
        positions = []
        for obj in objects:
            positions.append(Vector3D(*cmds.xform(obj, q=True, ws=True, rp=True)))
        centerPoint = Vector3D.findCenterPoint(positions)
        zgrp = BroNode.create(groupType, name=numberedName('{}_zgrp'.format(objects[0].name)))
        zgrp.setTranslation(centerPoint, space='world')
        if parent:
            for obj in objects:
                obj.parent = zgrp
    cmds.select(selection)

def setRGBColor(obj, color=(1, 1, 1)):
    ctrl = ensureBroNode(obj)
    rgb = ('R', 'G', 'B')
    ctrl.overrideEnabled = 1
    ctrl.overrideRGBColors = 1
    for (channel, color) in zip(rgb, color):
        cmds.setAttr((ctrl + ('.overrideColor%s' % channel)), color)

def getMaxBoundingBoxSize(obj, relative=True):
    log.debug(obj)
    if isinstance(obj, list):
        bbs = []
        for o in obj:
            log.debug(o)
            if isinstance(o, list):
                raise BroToolsError('getMaxBoundingBoxSize does not support nested lists!')
            bbs.append(getMaxBoundingBoxSize(o))
        max_size = mean(bbs)
    else:
        bb = cmds.xform(obj, q=True, bb=True)
        x = abs((bb[0] - bb[3]))
        y = abs((bb[1] - bb[4]))
        z = abs((bb[2] - bb[5]))
        max_size = max(x, y, z)
    if relative:
        max_size = round((max_size / 2.21638837511), 3)
    return max_size

def listParents(node):
    parents = cmds.ls(node, long=True)[0].split('|')[1:(- 1)]
    parents_long_named = ['|'.join(parents[:i]) for i in range(1, (1 + len(parents)))]
    return parents_long_named

def checkCreateGroup(name):
    "\n    Check if group exists and returns it's name if it does, creates if it does not.\n    :param name:\n    :return:\n    "
    if (cmds.objExists(name) != True):
        cmds.group(em=1, n=name)
    return name

def numberedName(name, zfill=3, fillType='number'):
    found = False
    i = 0
    new_name = name
    while (not found):
        if cmds.objExists(new_name):
            i = (i + 1)
            new_name = ((name + '_') + str(i).zfill(zfill))
        else:
            found = True
            return new_name

def numberedNameInData(name, data=None, zfill=3, fillType='number'):
    if (data == None):
        log.warning('numberedNameInData naming error! No data passed.', exc_info=True)
        return
    i = 0
    found = False
    new_name = name
    while (not found):
        if (new_name in data):
            i = (i + 1)
            new_name = ((name + '_') + str(i).zfill(zfill))
        else:
            found = True
            return new_name

def addLocatorAtSelectedObjects(objects=None, scale=None, *args):
    '\n    Adds a locator aligned to selected objects.\n    '
    if (objects == None):
        objects = cmds.ls(sl=1, l=1)
    if (scale == None):
        log.debug('addLocatorAtSelectedObjects round(getMaxBoundingBoxSize(objects), 2)', objects)
        scale = round(getMaxBoundingBoxSize(objects), 2)
    verts = cmds.filterExpand(sm=31)
    if (verts != None):
        objects.extend(verts)
    for obj in objects:
        objShortName = obj.split('|')[(- 1)]
        locator = cmds.spaceLocator(n=numberedName((objShortName + '_alignedLocator')))
        locatorShape = cmds.listRelatives(locator, s=True)[0]
        for axis in ['X', 'Y', 'Z']:
            cmds.setAttr(((locatorShape + '.localScale') + axis), scale)
        try:
            matchTransform(locator, obj, t=True, r=True)
        except:
            log.warning('Unable to match locator to', obj, 'deleting.')
            cmds.delete(locator)

def getDistance_positions(pos1, pos2):
    '\n        Get distance between two points in space.\n        Args:\n            obj1(str): Point 1\n            obj2(str): Points 2\n\n        Returns(float): Distance\n\n        '
    (Ax, Ay, Az) = pos1
    (Bx, By, Bz) = pos2
    dist = (((((Ax - Bx) ** 2) + ((Ay - By) ** 2)) + ((Az - Bz) ** 2)) ** 0.5)
    log.debug('DISTANCE between {} and {}: {}'.format(pos1, pos2, dist))
    return dist

def getDistance(obj1, obj2):
    '\n    Get distance between two objects.\n    Args:\n        obj1(str): Obj 1\n        obj2(str): Obj 2\n\n    Returns(float): Distance\n\n    '
    log.debug('Getting distance between {} and {}...'.format(str(obj1), str(obj2)))
    return getDistance_positions(cmds.xform(obj1, q=True, ws=True, rp=True), cmds.xform(obj2, q=True, ws=True, rp=True))

def getDistanceAverage(objects):
    prev_obj = None
    distance_sum = 0
    distance_count = 0
    for obj in objects:
        if (prev_obj == None):
            prev_obj = obj
            continue
        distance_sum += getDistance(prev_obj, obj)
        distance_count += 1
        prev_obj = obj
    avg_distance = (distance_sum / distance_count)
    return avg_distance

def getClosest(obj, objects):
    closest_obj = None
    closest_dist = None
    for object in objects:
        if (closest_dist == None):
            closest_dist = getDistance(obj, object)
            closest_obj = object
        elif (getDistance(obj, object) < closest_dist):
            closest_dist = getDistance(obj, object)
            closest_obj = object
    return closest_obj

def unlockAttributes(objects=None, attrs=['t', 'r'], axises=['x', 'y', 'z']):
    if (not issubclass(type(objects), list)):
        objects = [objects]
    if (objects == None):
        objects = cmds.ls(sl=1, l=1)
    for obj in objects:
        for attr in attrs:
            for axis in axises:
                cmds.setAttr((((obj + '.') + attr) + axis), lock=0, keyable=1)

def lockAttributes(objects=None, attrs=['t', 'r'], axises=['x', 'y', 'z']):
    if (not issubclass(type(objects), list)):
        objects = [objects]
    if (objects == None):
        objects = cmds.ls(sl=1, l=1)
    for obj in objects:
        for attr in attrs:
            for axis in axises:
                cmds.setAttr((((obj + '.') + attr) + axis), lock=1, keyable=1)

def getWSPosition(object, rp=False):
    if rp:
        return cmds.xform(object, q=1, ws=1, rp=1)
    else:
        return cmds.xform(object, q=1, ws=1, t=1)

def setWSPosition(object, pos, rp=False):
    if rp:
        return cmds.xform(object, ws=1, rp=pos)
    else:
        return cmds.xform(object, ws=1, t=pos)

def getWSRotation(object, rp=False):
    if rp:
        rotation = cmds.xform(object, q=1, ws=1, rp=1)
    else:
        rotation = cmds.xform(object, q=1, ws=1, ro=1)
    return rotation
try:
    gMainProgressBar = mel.eval('$tmp = $gMainProgressBar')
    log.debug('gMainProgressBar created.')
except:
    log.warning('Unable to create gMainProgressBar in BroTools.core.utils!', traceback=True)

def progressBarStart(maxValue, status='Working...'):
    cmds.progressBar(gMainProgressBar, edit=True, beginProgress=True, isInterruptable=True, status=status, maxValue=maxValue)
    cmds.progressBar(gMainProgressBar, e=True, maxValue=maxValue, pr=0, status=status)

def progressBarIsCanceled():
    return cmds.progressBar(gMainProgressBar, query=True, isCancelled=True)

def progressBarStep(step=1, status=None):
    if (status == None):
        cmds.progressBar(gMainProgressBar, edit=True, step=step)
    else:
        cmds.progressBar(gMainProgressBar, edit=True, step=step, status=status)

def progressBarEnd():
    cmds.progressBar(gMainProgressBar, edit=True, endProgress=True)

def bake_pivot(objects, pos=1, ori=1):
    cmds.select(objects, r=True)
    objects = cmds.ls(sl=1, transforms=1)
    shapes = cmds.ls(sl=1, shapes=1)
    if (len(shapes) > 0):
        transforms = cmds.listRelatives(path=1, parent=1, type='transform')
        objects += transforms
    if (len(objects) == 0):
        cmds.error('m_bakeCustomToolPivot.kNoObjectsSelectedError')
        return
    currentCtx = cmds.currentCtx()
    contextList = ['moveSuperContext', 'manipMoveContext', 'RotateSuperContext', 'manipRotateContext', 'scaleSuperContext', 'manipScaleContext']
    if (currentCtx not in contextList):
        cmds.error('m_bakeCustomToolPivot.kWrongToolError')
        return
    customOri = []
    pivotModeActive = 0
    customModeActive = 0
    if ((currentCtx == 'moveSuperContext') or (currentCtx == 'manipMoveContext')):
        customOri = cmds.manipMoveContext('Move', q=1, orientAxes=1)
        pivotModeActive = cmds.manipMoveContext('Move', q=1, editPivotMode=1)
        customModeActive = (cmds.manipMoveContext('Move', q=1, mode=1) / 6)
    elif ((currentCtx == 'RotateSuperContext') or (currentCtx == 'manipRotateContext')):
        customOri = cmds.manipRotateContext('Rotate', q=1, orientAxes=1)
        pivotModeActive = cmds.manipRotateContext('Rotate', q=1, editPivotMode=1)
        customModeActive = (cmds.manipRotateContext('Rotate', q=1, mode=1) / 3)
    elif ((currentCtx == 'scaleSuperContext') or (currentCtx == 'manipScaleContext')):
        customOri = cmds.manipScaleContext('Scale', q=1, orientAxes=1)
        pivotModeActive = cmds.manipScaleContext('Scale', q=1, editPivotMode=1)
        customModeActive = (cmds.manipScaleContext('Scale', q=1, mode=1) / 6)
    if (ori and (not pos) and (not customModeActive)):
        cmds.error('m_bakeCustomToolPivot.kWrongAxisOriModeError')
        return
    if (ori and customModeActive):
        customOri[0] = mel.eval('rad_to_deg({})'.format(customOri[0]))
        customOri[1] = mel.eval('rad_to_deg({})'.format(customOri[1]))
        customOri[2] = mel.eval('rad_to_deg({})'.format(customOri[2]))
        cmds.rotate(customOri[0], customOri[1], customOri[2], objects, a=1, pcp=1, pgp=1, ws=1, fo=1)
    if pos:
        for object in objects:
            old = [0, 0, 0]
            m = cmds.xform(object, q=1, m=1)
            p = cmds.xform(object, q=1, os=1, sp=1)
            old[0] = ((((p[0] * m[0]) + (p[1] * m[4])) + (p[2] * m[8])) + m[12])
            old[1] = ((((p[0] * m[1]) + (p[1] * m[5])) + (p[2] * m[9])) + m[13])
            old[2] = ((((p[0] * m[2]) + (p[1] * m[6])) + (p[2] * m[10])) + m[14])
            cmds.xform(object, zeroTransformPivots=1)
            new = cmds.getAttr((object + '.translate'))[0]
            cmds.move((old[0] - new[0]), (old[1] - new[1]), (old[2] - new[2]), object, pcp=1, pgp=1, ls=1, r=1)
    if pivotModeActive:
        mel.eval('ctxEditMode;')
    if (ori and customModeActive):
        if ((currentCtx == 'moveSuperContext') or (currentCtx == 'manipMoveContext')):
            cmds.manipMoveContext('Move', e=1, mode=0)
        elif ((currentCtx == 'RotateSuperContext') or (currentCtx == 'manipRotateContext')):
            cmds.manipRotateContext('Rotate', e=True, mode=0)
        elif ((currentCtx == 'scaleSuperContext') or (currentCtx == 'manipScaleContext')):
            cmds.manipScaleContext('Scale', e=1, mode=0)

def createAimLocators(axis=[1, 0, 0], up=[0, 1, 0], wuo='none', wu=[0, 1, 0], mo=False, wut='objectrotation'):
    log.info('>', 'AIM:', axis, up, wuo, wu, mo, wut)
    l1 = BroNode.create('spaceLocator', n=numberedName('bt_utils_aimLocator_l1'))
    l2 = BroNode.create('spaceLocator', n=numberedName('bt_utils_aimLocator_l2'))
    aim = cmds.aimConstraint(l2, l1, mo=mo, u=up, wut=wut, wuo=wuo, wu=up, aim=axis)
    return (l1, l2, aim)

def aimToPosition(object, targetPos, l1=None, l2=None, key=True):
    if ((l1 == None) or (l2 == None)):
        aimTransform(object, targetPos)
    else:
        setWSPosition(l1, getWSPosition(object, rp=True))
        setWSPosition(l2, targetPos)
        matchTransform(object, l1, t=False, r=True)
        cmds.select(cl=True)
    if key:
        cmds.setKeyframe(object)

def setWSRotation(object, rot):
    return cmds.xform(object, ws=1, ro=rot)

def buildNurbsShapesForJoints(selection, fwdAxis):
    for (i, jnt) in enumerate(selection):
        log.info('JNT', jnt)
        try:
            mv = True
            log.info('creating nurbs shape with distance')
            dist = getDistance(jnt, selection[(i + 1)])
            if (fwdAxis == 'X'):
                nurbs = cmds.nurbsCube(w=1, hr=dist, lr=1, ch=False)
            elif (fwdAxis == 'Y'):
                nurbs = cmds.nurbsCube(w=1, hr=1, lr=dist, ch=False)
            elif (fwdAxis == 'Z'):
                nurbs = cmds.nurbsCube(w=dist, hr=0.1, lr=0.1, ch=False)
        except:
            mv = False
            log.info('Failed, creating just nurbs shape')
            nurbs = cmds.nurbsCube(hr=2, lr=2, w=2, ch=False)
        shapes = cmds.listRelatives(nurbs[0], ad=True)
        log.info(nurbs[0], shapes)
        for shape in shapes:
            cmds.select(shape, r=True)
            cmds.nurbsSelect()
            if (mv == True):
                if (fwdAxis == 'X'):
                    cmds.move((dist / 2), 0, 0, r=True, os=True, wd=True)
                elif (fwdAxis == 'Y'):
                    cmds.move(0, (dist / 2), 0, r=True, os=True, wd=True)
                elif (fwdAxis == 'Z'):
                    cmds.move(0, 0, (dist / 2), r=True, os=True, wd=True)
            log.info('parenting shape', shape)
            try:
                cmds.parent(shape, jnt, r=True, s=True)
            except:
                pass
        cmds.setAttr((jnt + '.drawStyle'), 2)

def buildNurbsShapesForJointsInverted(selection, fwdAxis):
    for (i, jnt) in enumerate(selection):
        print('JNT', jnt)
        try:
            mv = True
            print('creating nurbs shape with distance')
            dist = getDistance(jnt, selection[(i + 1)])
            if (fwdAxis == 'X'):
                nurbs = cmds.nurbsCube(w=1, hr=dist, lr=1, ch=False)
            elif (fwdAxis == 'Y'):
                nurbs = cmds.nurbsCube(w=1, hr=1, lr=dist, ch=False)
            elif (fwdAxis == 'Z'):
                nurbs = cmds.nurbsCube(w=dist, hr=0.1, lr=0.1, ch=False)
        except:
            mv = False
            print('Failed, creating just nurbs shape')
            nurbs = cmds.nurbsCube(hr=2, lr=2, w=2, ch=False)
        shapes = cmds.listRelatives(nurbs[0], ad=True)
        print(nurbs[0], shapes)
        for shape in shapes:
            cmds.select(shape, r=True)
            cmds.nurbsSelect()
            if (mv == True):
                if (fwdAxis == 'X'):
                    cmds.move(((dist / 2) * (- 1)), 0, 0, r=True, os=True, wd=True)
                elif (fwdAxis == 'Y'):
                    cmds.move(0, ((dist / 2) * (- 1)), 0, r=True, os=True, wd=True)
                elif (fwdAxis == 'Z'):
                    cmds.move(0, 0, ((dist / 2) * (- 1)), r=True, os=True, wd=True)
            print('parenting shape', shape)
            try:
                cmds.parent(shape, jnt, r=True, s=True)
            except:
                pass
        cmds.setAttr((jnt + '.drawStyle'), 2)

def selectFilter(oType):
    selection = cmds.ls(sl=True)
    joints = []
    for obj in selection:
        if cmds.objectType(obj, i=oType):
            joints.append(obj)
    cmds.select(cl=True)
    for obj in joints:
        cmds.select(obj, add=True)

def zeroGroupNodes(nodes, suffix='_zGRP', removeSuffix=False):
    log.info('', 'Creating zero group.')
    newGroups = []
    for obj in nodes:
        print('Grouping', obj, 'suffix', suffix)
        objName = obj.split('|')
        objName = objName[(len(objName) - 1)]
        parentObj = cmds.listRelatives(obj, parent=True)
        print('Parent', parentObj)
        if removeSuffix:
            objNameSplit = objName.split('_')
            zGRPName = objName.replace(('_' + objNameSplit[(len(objNameSplit) - 1)]), suffix)
        else:
            zGRPName = (objName + suffix)
        zGRP = cmds.group(n=zGRPName, em=True, w=True)
        matPos = cmds.xform(obj, q=True, ws=True, m=True)
        cmds.xform(zGRP, ws=True, m=matPos)
        cmds.parent(obj, zGRP)
        if (parentObj != None):
            cmds.parent(zGRP, parentObj)
        newGroups.append(zGRP)
    log.info('', 'Done')
    return newGroups

def playblast_panel_accepted(panel):
    accepted = False
    if (panel != None):
        if (cmds.getPanel(to=panel) == 'modelPanel'):
            accepted = True
        else:
            log.debug(panel, 'is not modelPanel')
    else:
        log.debug(panel, 'is None')
    return accepted

def makeQuickPlayblast(panel=None, camera='', ref='', msaa=True, hwFogEnable=True, hwFogDensity=0.0003, hwFogFalloff=2, hwFogColor=[0.0761852, 0.0761852, 0.0761852], hwFogAlpha=1, ssaoEnable=True, shadows=True, da='smoothShaded', lights=True, textures=True, playblast=True, saveScene=False, filename=None, frames_in_filename=False, widthHeight=[1920, 1080], film_gate=True):
    fname = 'makeQuickPlayblast'
    if (ref != ''):
        log.info(fname, 'referencing file:', ref)
        cmds.file(ref, r=True, dns=True)
    if (not panel):
        panel = cmds.getPanel(up=True)
        if (not playblast_panel_accepted(panel)):
            panel = cmds.getPanel(wf=True)
        if (not playblast_panel_accepted(panel)):
            visible_panels = cmds.getPanel(vis=True)
            for p in visible_panels:
                if playblast_panel_accepted(p):
                    panel = p
    if (not playblast_panel_accepted(panel)):
        raise BroToolsError("No panel could be accepted or found that's suitable for playblasting.")
    if playblast_panel_accepted(panel):
        log.info(panel, 'accepted for playblasting.')
        user_cam = cmds.modelEditor(panel, q=True, camera=True)
        cmds.modelEditor(panel, e=True, rnm='vp2Renderer', backfaceCulling=1)
        cmds.modelEditor(panel, e=True, allObjects=False)
        cmds.modelEditor(panel, e=True, imagePlane=True)
        cmds.modelEditor(panel, e=True, polymeshes=True)
        cmds.modelEditor(panel, e=True, lights=False)
        cmds.modelEditor(panel, e=True, nParticles=True)
        cmds.setAttr('hardwareRenderingGlobals.msaa', msaa)
        cmds.setAttr('hardwareRenderingGlobals.ssaoEnable', ssaoEnable)
        log.debug(cmds.modelEditor(panel, q=True, rls=True))
        cmds.modelEditor(panel, e=True, shadows=shadows, da=da, textures=textures)
        if cmds.objExists(camera):
            cmds.lookThru(camera)
            cmds.camera(camera, e=True, dfg=film_gate)
        else:
            log.debug('!!! CAMERA NOT FOUND:', camera)
        sceneName = ((os.path.splitext(os.path.basename(cmds.file(q=True, sn=True)))[0] + '_') + BroNode(camera).name)
        if frames_in_filename:
            sceneName += '_s{}-e{}'.format(int(cmds.playbackOptions(q=True, min=True)), int(cmds.playbackOptions(q=True, max=True)))
        playblast_filename = None
        if playblast:
            cmds.modelEditor(panel, e=True, dtx=False)
            cmds.refresh(f=True)
            cmds.modelEditor(panel, e=True, dtx=True)
            for i in range(0, 10):
                log.debug('Refresh', i)
                cmds.refresh()
            if (filename == None):
                folderpath = os.path.join(cmds.workspace(q=True, rd=True), 'movies')
                if (not os.path.exists(folderpath)):
                    os.makedirs(folderpath)
                filename = os.path.normpath(os.path.join(folderpath, sceneName))
            log.info('Recording playblast into: {}'.format(filename))
            playblast_filename = cmds.playblast(format='movie', compression='IYUV codec', filename=filename, showOrnaments=0, percent=100, quality=92, widthHeight=widthHeight, fo=True, viewer=False, os=False)
            log.info('Playblast recorded as: {}'.format(playblast_filename))
        log.debug('Return camera to', user_cam)
        cmds.lookThru(user_cam)
        cmds.setAttr('hardwareRenderingGlobals.msaa', False)
        cmds.setAttr('hardwareRenderingGlobals.ssaoEnable', False)
        return (playblast_filename + '.avi')

def makeQuickPlayblast_obsolete(panel='none', camera='', ref='', msaa=True, hwFogEnable=True, hwFogDensity=0.0003, hwFogFalloff=2, hwFogColor=[0.0761852, 0.0761852, 0.0761852], hwFogAlpha=1, ssaoEnable=True, shadows=True, da='smoothShaded', lights=True, textures=True, playblast=True):
    fname = 'makeQuickPlayblast'
    if (ref != ''):
        print(fname, 'referencing file:', ref)
        cmds.file(ref, r=True, dns=True)
    cmds.modelEditor(panel, e=True, rnm='vp2Renderer', backfaceCulling=1)
    if (panel == 'none'):
        panel = cmds.getPanel(wf=True)
    cmds.modelEditor(panel, e=True, alo=False)
    cmds.modelEditor(panel, e=True, pm=True)
    cmds.setAttr('hardwareRenderingGlobals.msaa', msaa)
    cmds.setAttr('hardwareRenderingGlobals.ssaoEnable', ssaoEnable)
    cmds.setAttr('hardwareRenderingGlobals.hwFogEnable', hwFogEnable)
    cmds.setAttr('hardwareRenderingGlobals.hwFogDensity', hwFogDensity)
    cmds.setAttr('hardwareRenderingGlobals.hwFogFalloff', hwFogFalloff)
    cmds.setAttr('hardwareRenderingGlobals.hwFogColor', hwFogColor[0], hwFogColor[1], hwFogColor[2], type='double3')
    cmds.setAttr('hardwareRenderingGlobals.hwFogAlpha', hwFogAlpha)
    print(cmds.modelEditor(panel, q=True, rls=True))
    cmds.modelEditor(panel, e=True, shadows=shadows, da=da, lights=lights, textures=textures)
    if cmds.objExists(camera):
        cmds.lookThru(camera)
        cmds.camera(camera, e=True, dfg=True)
    else:
        print('!!! CAMERA NOT FOUND:', camera)
    sceneName = cmds.file(q=True, sceneName=True, shn=True).replace('.mb', '')
    if playblast:
        print('PLAYBLASTING...')
        cmds.playblast(format='qt', compression='JPEG', filename=(('movies/' + sceneName) + '.mov'), showOrnaments=0, percent=100, quality=92, widthHeight=[1920, 1080], fo=True)
    "\n    import os\n\n    ref = r'E:\\PersonalWork\\ShowReels\\MayaStuff\\previewScene.ma'\n    folder = 'E:\\Work\\FOX3D\\BattleSouls\\RigsAnims\\ShadowMage\\Anims\\'\n\n    for root, dirs, files in os.walk(folder, topdown=False):\n        for name in files:\n            filePath = os.path.join(root,name)\n            n = name.split('.')[0]\n            if '.ma' in filePath and 'ShadowMage_' in filePath and '.mayaSwatches' not in filePath:\n                print filePath, n\n                cmds.file (filePath, open=True, force=True)\n                BroTools.core.commutils.makeQuickPlayblast(panel=u'modelPantel4', camera='PREVIEW_CAM', ref=ref)\n    "
    '\n    from shutil import copyfile\n\n    folder = r\'E:\\Work\\FOX3D\\LooneyTunes\\Animation\\Bugs_Bunny_Kiss\\Bugs_Bunny\\scenes\\\'\n    animNames = [\'BugsBunny_curious.0002\', \'BugsBunny_Default_Intro\', \'BugsBunny_Disgusted.0002\', \'BugsBunny_idle.0002\', \'BugsBunny_Idle_fish.0002\', \'BugsBunny_In_Love.0002\', \'BugsBunny_Kiss_Fail.0001\', \'BugsBunny_Kiss_Success\', \'BugsBunny_Kissed.0003\', \'BugsBunny_Kissed_fail_fish.0004\', \'BugsBunny_Sexy_Intro.0001\', \'BugsBunny_Sexy_Move.0002\', \'BugsBunny_Sexy_Walk.0002\']\n\n    for anim in animNames:\n        filePath = folder+anim+\'.mb\'\n        copyfile(filePath, filePath.replace(\'scenes\',\'sendFiles\'))\n        print "Working with", filePath\n        cmds.file (filePath, open=True, force=True)\n\n        BroTools.core.commutils.makeQuickPlayblast(panel=u\'modelPanel4\')\n\n    '

def unlock_attrs(obj=None, *args):
    if (obj == None):
        obj = cmds.ls(sl=0)
    for a in ['t', 'r', 's']:
        for x in ['x', 'y', 'z']:
            cmds.setAttr((((obj + '.') + a) + x), e=True, l=False)

def extract_blendshapes(*args):
    '\n    Extracts blendshapes from the selected mesh into a group named BS. Extracted meshes have cleaned CH.\n    :param args:\n    :return:\n    '
    obj = cmds.ls(sl=1)[0]
    bs = cmds.ls(cmds.listHistory(obj), type='blendShape')[0]
    cmds.select(cl=True)
    grp = cmds.group(n='BS', em=True)
    attrs = cmds.listAttr((bs + '.w'), m=True)
    extracted = []
    gMainProgressBar = mel.eval('$tmp = $gMainProgressBar')
    cmds.progressBar(gMainProgressBar, edit=True, beginProgress=True, isInterruptable=True, status='Processing...', maxValue=len(attrs))
    for (i, attr) in enumerate(attrs):
        if cmds.progressBar(gMainProgressBar, query=True, isCancelled=True):
            break
        for attr_2 in attrs:
            cmds.setAttr(((bs + '.') + attr_2), 0)
        cmds.setAttr(((bs + '.') + attr), 1)
        bsObj = cmds.duplicate(obj)
        bsObj = cmds.rename(bsObj, attr)
        cmds.delete(bsObj, ch=True)
        cmds.parent(bsObj, grp)
        cmds.setAttr((bsObj + '.visibility'), 0)
        extracted.append(bsObj)
        cmds.progressBar(gMainProgressBar, edit=True, step=1)
    cmds.progressBar(gMainProgressBar, edit=True, endProgress=True)
    return (obj, extracted)

def extract_blendshapes_sl(*args):
    print('Extracting blendshapes...')
    (obj, extracted) = extract_blendshapes()
    cmds.delete(obj, ch=True)
    extracted.append(obj)
    for a in extracted:
        unlock_attrs(a)
    cmds.select(extracted, r=True)
    print('Cleaning up...')
    import bt_zeroTransforms
    bt_zeroTransforms.run(mode='origin')
    print('Done.')

def axis_array_to_letter(arr):
    '\n    Takes in array and returns appropriate axis letter - x, y, z\n    '
    if ((arr == [1, 0, 0]) or (arr == [(- 1), 0, 0])):
        return 'x'
    elif ((arr == [0, 1, 0]) or (arr == [0, (- 1), 0])):
        return 'y'
    elif ((arr == [0, 0, 1]) or (arr == [0, 0, (- 1)])):
        return 'z'
    else:
        raise BroToolsError('Cant determine axis from: {}'.format(str(arr)))
        return

def axis_letter_to_array(axis, negative=False):
    '\n    Takes in axis letter and returns appropriate axis array\n    '
    if (axis == 'x'):
        if negative:
            return [(- 1), 0, 0]
        return [1, 0, 0]
    elif (axis == 'y'):
        if negative:
            return [0, (- 1), 0]
        return [0, 1, 0]
    elif (axis == 'z'):
        if negative:
            return [0, 0, (- 1)]
        return [0, 0, 1]
    else:
        raise BroToolsError('Cant determine axis array from: {}'.format(str(axis)))

def addObjectsToAnimLayer(objects, animLayer):
    ' Add the specified objects to the specified animation layer. '
    if (not hasattr(objects, '__iter__')):
        objects = [objects]
    for obj in objects:
        attributes = cmds.listAttr(obj, keyable=True, unlocked=True)
        if (not attributes):
            log.warning('Object "{0}" had no keyable attributes to add to animation layer.')
            continue
        nodeAttrs = ['{0}.{1}'.format(obj, attr) for attr in attributes]
        cmds.animLayer(animLayer, e=True, attribute=nodeAttrs)

def get_cv_points(obj):
    "\n    Get shape data compatible with BroRig's shapes\n    "
    try:
        c = BroNode(obj)
        shp = BroNode(cmds.listRelatives(c, shapes=True)[0])
        degs = shp.degree
        spans = shp.spans
        form = shp.form
        cvs = cmds.getAttr('{}.cv[*]'.format(shp))
        curveInfo = BroNode.create('curveInfo')
        curveInfo.connect(c, 'worldSpace[0]', 'inputCurve')
        knots = curveInfo.knots
        if knots:
            knots = knots[0]
        points = [degs, form, knots]
        points.extend(cvs)
        curveInfo.delete()
        return points
    except Exception as e:
        raise BroToolsError(e, silent=False)
get_shape_cv_points = get_cv_points

def nodesToLongNames(nodes):
    result = []
    for node in nodes:
        try:
            s = node.longName()
        except:
            try:
                s = node.longName
            except:
                s = str(node)
        result.append(s)
    return result

def namesToBroNodes(names):
    result = []
    for name in names:
        result.append(BroNode(name))
    return result
log.debug('Import inspect')
import inspect
log.debug('Import ModuleType')
from types import ModuleType
log.debug('Import complete')

def get_default_args(func):
    '\n    returns a dictionary of arg_name:default_values for the input function\n    '
    (args, varargs, keywords, defaults) = inspect.getargspec(func)
    return dict(list(zip(reversed(args), reversed(defaults))))
reloaded_modules = []
ignored_modules = []
reload_errors = []

def rreload_old(module, depth=2, top=True, reload_names=[]):
    'Recursively reload modules.'
    if top:
        log.info('rreload module', module, 'started...')
    if (depth < 1):
        return
    reload(module)
    for attribute_name in dir(module):
        if ((attribute_name != 'BroTools') and (attribute_name != 'plug-ins') and (attribute_name in reload_names)):
            try:
                attribute = getattr(module, attribute_name)
                if (type(attribute) is ModuleType):
                    reloaded_modules.append(attribute_name)
                    rreload(attribute, depth=(depth - 1), top=False)
            except:
                reload_errors.append([attribute_name, traceback.format_exc()])
        else:
            ignored_modules.append(attribute_name)
    if top:
        log.info(reloaded_modules)
        log.info(ignored_modules)
        log.info(reload_errors)
        log.info('rreload finished.')
log.debug('from BroTools import bro_reload')
from BroTools import bro_reload

def rreload(*args, **kwargs):
    bro_reload.doReload()

def get_all_modules():
    bro_modules = []
    "\n    # OPTION 1\n    recursive_files = [os.path.join(dp, f) for dp, dn, fn in os.walk(os.path.expanduser(rp)) for f in fn]\n    for file in recursive_files:\n        if os.path.isdir(file):\n            bro_modules.append(os.path.basename(file))\n        else:\n            if '.py' in file and '__init__' not in file:\n                bro_modules.append(os.path.basename(file.replace('.py', '')))\n    "
    for (root, dirs, files) in os.walk(modpath, topdown=True):
        for name in files:
            if ('.py' in name.lower()):
                bro_modules.append(name.replace('.py', '').replace('.PY', ''))
        for name in dirs:
            bro_modules.append(name)
    return bro_modules

class brokwargs():
    '\n    replacer for kwargs.get, supporting long and short names.\n    '

    def __init__(self, labelLong, labelShort, defautValue, kwargDic=None):
        if (kwargDic == None):
            raise BroToolsError('You forgot to pass kwargs to brokwargs!')
        if (kwargDic.get(labelLong) and kwargDic.get(labelShort)):
            print('Same flag used twice.')
        else:
            self.userInput = (kwargDic.get(labelLong) or kwargDic.get(labelShort) or defautValue)

    def get(self):
        return self.userInput

def recreate_hierarchy_simple(jnt, p=None):
    '\n    Simple function to recreate hierarchy of joints\n    '
    cmds.select(cl=True)
    nj = cmds.joint(n=(jnt + '_new'))
    cmds.matchTransform(nj, jnt)
    if (p != None):
        nj = cmds.parent(nj, p)
    children = cmds.listRelatives(jnt, children=True)
    if (children != None):
        for j in children:
            recreate_hierarchy_simple(j, nj)

def randomHash(length=6, filter=[], prefix='BTRnd_'):
    '\n\n    :param length: Number of symbols in hash\n    :param filter: Duplicate filter, can be list or dict\n    :return:\n    '
    hs = str(uuid.uuid4().hex[:length].upper())
    while (hs in filter):
        hs = str(uuid.uuid4().hex[:length].upper())
    return (prefix + hs)

def zp(data):
    try:
        r = base64.b64encode(zlib.compress(data))
        return r
    except:
        return 'Err'

def isDigit(x):
    try:
        float(x)
        return True
    except ValueError:
        return False

def isInteger(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

def value_from_string(value):
    if (value.strip() in ['True', 'true']):
        return True
    elif (value.strip() in ['False', 'false']):
        return False
    if isInteger(value):
        return int(value)
    elif isDigit(value):
        return float(value)
    else:
        return str(value)

def modules_in_dir(dir):
    '\n    This function will find any modules in the given folder, and return them as a string ready for use with importlib\n\n    :param dir: Directory to scan for modules, relative to the root of the project.\n    :return: A list of module paths relative to the root of the project, ready to use with importlib.\n    '
    path = os.path.join(modpath, dir)
    files = os.listdir(os.path.join(modpath, dir))
    modules = []
    for file in files:
        (name, ext) = os.path.splitext(file)
        if (ext.lower() in ['.py']):
            file_path = os.path.join(path, file)
            mod_path = file_path[file_path.find(dir):]
            mod_path = mod_path.replace(ext, '').replace(os.sep, '.')
            modules.append(mod_path)
    return modules

def open_web_url(*args, **kwargs):
    if ('url' in kwargs):
        url = kwargs.get('url')
    else:
        url = args[0]
    try:
        webbrowser.open(url)
    except Exception as e:
        raise BroToolsError('Tried to open web-link ({}; {}; {}), but experienced an error: {}'.format(args[0], args, kwargs, e), exc_info=True)

def undo(fn):
    log.error('Note to Developer: Use Catch Errors instead of undo')
from BroTools.core import suffixes
known_suffixes = []
for (k, v) in suffixes.__dict__.items():
    if ('_sfx' in k):
        if (isinstance(v, str) or isinstance(v, unicode)):
            known_suffixes.append(v)
log.debug('Known suffixes: {}'.format(known_suffixes))

def changeSuffix(name, new_suffix, smart=False):
    new_suffix = str(new_suffix)
    (clear_name, separator, existing_suffix) = name.rpartition('_')
    if ((clear_name == '') and (separator == '')):
        new_name = ((name + '_') + new_suffix)
        return new_name
    if smart:
        if (existing_suffix in known_suffixes):
            new_name = ((clear_name + '_') + new_suffix)
        else:
            new_name = ((name + '_') + new_suffix)
    else:
        new_name = ((clear_name + '_') + new_suffix)
    return new_name
shorten_data = {'pointConstraint': 'pos', 'orientConstraint': 'orient', 'scaleConstraint': 'scale', 'parentConstraint': 'parent'}

@catch_errors
def shorten(name):
    if (name in shorten_data):
        return shorten_data[name]
    else:
        return camel_case_split(name)[0].split('_')[0].split(' ')[0][0:5]

def get_scene_number_from_name(name):
    pattern = '.*(?P<number>\\d\\d\\d\\d)\\...'
    number = None
    m = re.match(pattern, name)
    if m:
        number = m.groupdict().get('number', None)
    return number

def strip_number_from_scene_name(name):
    number = get_scene_number_from_name(name)
    if (number is None):
        return name
    else:
        return name.replace(('.' + number), '')

def getKeyframeTimes(objects=None, integers=False, respect_timeline_selection=True):
    '\n    Returns all frames with keyframes on them, only once.\n    If no objects passed, will select all transforms.\n    '
    if (objects is None):
        objects = cmds.ls(type='transform')
        objects.extend(cmds.ls(type='joint'))
        objects.extend(cmds.ls(type='nurbsCurve'))
    if respect_timeline_selection:
        (start, end) = getCurrentTimeRange()
    keyframes = []
    for obj in objects:
        keyframe_data = cmds.keyframe(obj, query=True, timeChange=True)
        if keyframe_data:
            for i in keyframe_data:
                if ((respect_timeline_selection is False) or (start <= i <= end)):
                    if (i not in keyframes):
                        if integers:
                            i = int(i)
                        keyframes.append(i)
        else:
            log.debug('No keyframes found on: {}'.format(obj))
    return keyframes

def cropTimelineToAnimation(*args):
    print('cropTimelineToAnimation', args)
    keyframe_times = getKeyframeTimes(respect_timeline_selection=False)
    animStart = min(keyframe_times)
    animEnd = max(keyframe_times)
    cmds.playbackOptions(min=animStart, max=animEnd, ast=animStart, aet=animEnd)
    print('Timeline set. Start:', animStart, 'end:', animEnd)

def keep_selection(fn):
    '\n    Keep user selection\n    '

    def wrapped(*args, **kwargs):
        s = cmds.ls(sl=True, long=True)
        log.debug('Keep user selection: {}'.format(s))
        result = fn(*args, **kwargs)
        log.debug('Restore user selection: {}'.format(s))
        cmds.ls(s, r=True)
        return result
    return wrapped
log.debug('BroTools.core.maya_utils end reached')
