from base64 import b64decode as license64

'\nThis script is used to gather some information to help diagnose less common problems with BroTools.\nIt will collect information about your system environment such as OS, screen resolution and similar and log (it out\nin the script editor window. It will then try to import various modules and help determine which may be causing the problem\n\nYou may edit your system information out if you find it sensitive, but it is required for developer to be able to\nrecreate your environment as closely as possible inside a VM or otherwise.\n\nTo run it use following command:\n\nfrom BroTools import diagnostic\nreload(diagnostic)\ndiagnostic.run()\n'
__version__ = '2024.06.14'
import traceback
import os
import sys
import locale
import hashlib
import platform
import subprocess
from datetime import datetime
MOD_PATH = APP_DIR = os.path.dirname(os.path.realpath(__file__))
LOG_PATH = os.path.join(APP_DIR, 'diagnostic.log')
if os.path.exists(LOG_PATH):
    os.remove(LOG_PATH)

def log(*args, **kwargs):
    err = kwargs.get('err', False)
    text = ''
    for a in args:
        text += (str(a) + ' ')
    text += '\n'
    print(text)
    with open(LOG_PATH, 'a+') as f:
        f.write(text)
        if err:
            f.write(traceback.format_exc())
            print(('- ' * 10))
            traceback.print_exc()
            print(('- ' * 10))

def run():
    from maya import cmds, mel
    try:
        log(('# ' * 30))
        log(('Diagnostic module version:' + __version__))
        log('Date: {}'.format(datetime.now()))
        log(('# ' * 30))
        log('Gathering system information...')

        def brotools_hash():
            hash_files = []
            for (root, dirs, files) in os.walk(MOD_PATH, topdown=False):
                for filename in files:
                    (name, ext) = os.path.splitext(filename)
                    if (ext.lower() == '.py'):
                        filepath = os.path.join(root, filename)
                        relpath = os.path.relpath(filepath, MOD_PATH)
                        hash_files.append(filepath)
            hash_obj = hashlib.sha1(open(hash_files[0], 'rb').read())
            for fname in hash_files[1:]:
                hash_obj.update(open(fname, 'rb').read())
            checksum = hash_obj.hexdigest()
            return checksum

        def individual_files_hash():
            hash_files = []
            hash_data = []
            for (root, dirs, files) in os.walk(MOD_PATH, topdown=False):
                for filename in files:
                    (name, ext) = os.path.splitext(filename)
                    if (ext.lower() == '.py'):
                        filepath = os.path.join(root, filename)
                        relpath = os.path.relpath(filepath, MOD_PATH)
                        hash_files.append(filepath)
                        hash_data.append([relpath, file_hash(filepath)])
            return hash_data

        def file_hash(filepath):
            hash_obj = hashlib.sha1(open(filepath, 'rb').read())
            checksum = hash_obj.hexdigest()
            return checksum
        log('\n-----------------')
        log('-  {:<15} {}'.format('Date started:', str(datetime.now())))
        log('\nMAYA INFO')
        log('-  {:<15} {}'.format('OS:', cmds.about(q=True, os=True)))
        log('-  {:<15} {}'.format('UI Lang:', cmds.about(q=True, uil=True)))
        log('-  {:<15} {}'.format('Maya:', cmds.about(q=True, v=True)))
        log('-  {:<15} {}'.format('Maya full:', cmds.about(q=True, iv=True)))
        log('-  {:<15} {}'.format('Plugins:', cmds.pluginInfo(q=True, lsp=True)))
        log('-  {:<15} {}'.format('Python:', sys.version))
        log('\nSYSTEM INFO')
        log('-  {:<15} {}, {}'.format('OS:', platform.system(), platform.release()))
        log('-  {:<15} {}'.format('Locale Def:', locale.getdefaultlocale()))
        log('-  {:<15} {}'.format('Locale:', locale.getlocale()))
        log('\nADDITIONAL INFO')
        log('-  {:<15} {}'.format('MOD_PATH:', MOD_PATH))
        log('-  {:<15} {}'.format('brotools files hash:', individual_files_hash()))
        processes = []
        for process in subprocess.check_output('wmic process get description', shell=True).split('\n'):
            processes.append(process.strip().replace('\r', ''))
        log(('-  Running processes:' + str(processes)))
        if cmds.about(q=True, win=True):
            import ctypes
            user32 = ctypes.windll.user32
            screensize = (user32.GetSystemMetrics(0), user32.GetSystemMetrics(1), user32.GetSystemMetrics(78))
            log('-  {:<15} {}'.format('Screen Info:', str(screensize)))
            log('-  {:<15} {}'.format('Win:', str(sys.getwindowsversion())))
        log("\nIf you wish to keep any of this information private you're free to remove\n        some lines from here, however it can make debugging process harder. One part of\n        debugging process is to recreate your error. If I can't do that in my main system\n        my next step is to start a Virtual Machine with your OS and environment as close to\n        yours as possible. That's what all the above data is used for. \n")
        log('-----------------\n')
        log('Now trying to start importing various modules and attempting various actions. If Maya hangs after one of them - send screenshot of Script Editor output to developer. Last action that this script tried to perform is likely causing problems.')
        log('Trying to import PySide2.QtWebEngineWidgets ')
        try:
            import PySide2.QtWebEngineWidgets
        except:
            log("Failed to import PySide2.QtWebEngineWidgets, but seems like Maya did not hang. Most likely it's pre-2017 version of Maya that uses PySide1, which should not be causing problems.", err=True)
        log("Now trying to load bullet.mll plugin. If Maya hangs after this - it's likely your GPU is not supported by Maya's Bullet physics engine. If it happens, follow this steps to disable RBD module in BroDynamics: https://michaeldavydov.com/BroTools/Documentation/BroDynamics/troubleshooting/#solution-2")
        log(cmds.loadPlugin('bullet'))
        log('Looks ok if we got here. No trying to import various modules.')
        import importlib
        modules = ['core.broqt.vendor.Qt', 'core.broqt.bro_widgets', 'core.broqt.bro_windows', 'core.broqt.widgets']
        for (root, dirs, files) in os.walk(MOD_PATH, topdown=False):
            for filename in files:
                (name, ext) = os.path.splitext(filename)
                if (ext.lower() == '.py'):
                    filepath = os.path.join(root, filename)
                    relpath = os.path.relpath(filepath, MOD_PATH)
                    modules.append(relpath.replace(os.path.sep, '.').replace('.py', ''))
        for module in modules:
            module = ('BroTools.' + module)
            command = 'python ("import {}")'.format(module)
            log('Trying to import module {} [{}]...'.format(module, command))
            try:
                mel.eval(command)
            except:
                log('Error importing {}'.format(module), err=True)
        log("Diagnostic finished. If it reached this stage it most likely did not find any problems. Further debugging is required. Please send it's log to developer. This tool might need improving to detect more issues, or developer may have to RDP into your machine to try and find the issue manually.")
    except:
        log('General exception occured.', err=True)
        cmds.error('General exception occured. Please send log to developer.')
run()
