__version__ = "2021.3.10"
__release_date__ = "2020-12-12 21:10:02.351000"

# region # # # # # # # # # # # # # # # # # # MOD VARIABLES # # # # # # # # # # # # # # # # # #


MOD_NAME = "BroTools"
MOD_Prefix = "Bro"

import sys

PY3 = sys.version_info[0] >= 3

if PY3 == True:
    import builtins

    builtins.unicode = str

    PackagePaths = [
        ["packages_3_3"],
        ["BroRig", "plugins"]
    ]
else:
        PackagePaths = [
        ["packages_2_7"],
        ["BroRig", "plugins"]
    ]
MOD_PATH = ""

# endregion


import codecs
import importlib
import os
import sys
import traceback
import hashlib
import inspect
import time
# region # # # # # # # # # # # # # # # # # # IMPORTS # # # # # # # # # # # # # # # # # #
from functools import partial, wraps
from types import ModuleType

import maya.OpenMayaMPx as OpenMayaMPx
import maya.cmds as cmds
import maya.mel as mel

# maya 2016 and before vs maya 2017 and after
try:
    from PySide import QtCore
except:
    from PySide2 import QtCore

# endregion

# ----------------------------------------------------------------------------------------------------------------------
# region GLOBAL VARS
# ----------------------------------------------------------------------------------------------------------------------
plugin_path = cmds.pluginInfo(MOD_NAME + '.py', query=True, path=True)
plugin_dir = os.path.dirname(plugin_path)

MOD_dir_relative = os.path.join(plugin_dir, '..', '..')
MOD_PATH = os.path.abspath(MOD_dir_relative)


def verbose_log(modpath):
    '''
    Self contained verbose log check.
    v 1.0.0
    '''
    try:
        brotools_write_path = os.environ.get("BROTOOLS_CONFIG_PATH")
        if not brotools_write_path:
            brotools_write_path = modpath
        config_path = os.path.join(brotools_write_path, "config", "brotools.cfg")
        if os.path.exists(config_path):
            with open(config_path, "r") as f:
                lines = f.readlines()
        else:
            lines = []
            print("CANT FIND brotools.cfg FILE AT", config_path)
        verbose = False

        for line in lines:
            if "debug_level: 0" in line or "debug_level:0" in line:
                verbose = True
                print("BROTOOLS.py VERBOSE LOG ENABLED")
        return verbose
    except Exception as e:
        raise Exception("{} verbose_log failure: {}".format(__file__, e))


VERBOSE_LOG = verbose_log

# endregion

brotools_path = os.path.join(MOD_PATH, MOD_NAME)

# Reroute maya log output to file
brotools_write_path = os.environ.get("BROTOOLS_CONFIG_PATH")
if not brotools_write_path:
    brotools_write_path = os.path.join(MOD_PATH, MOD_NAME)
log_path = os.path.join(brotools_write_path, 'bro_startup.log')

if os.path.exists(log_path):
    try:
        os.remove(log_path)
        print("Old log file removed", log_path)
    except:
        pass

cmdFileOutput_descriptor = cmds.cmdFileOutput(o=log_path)


def remove_pyc_files():
    log("Cleaning up .pyc files...")
    errors = False
    for root, dirs, files in os.walk(brotools_path, topdown=False):
        for filename in files:
            name, ext = os.path.splitext(filename)
            if ext.lower() == ".pyc":
                filepath = os.path.join(root, filename)
                try:
                    os.remove(filepath)
                except Exception as e:
                    errors = True
                    log("   Failed to cleanup .pyc file [{}], {}.".format(filepath, e), traceback=True)
    if not errors:
        log("   .pyc cleanup finished.")
    else:
        log("   .pyc cleanup finished with errors.")


def brotools_hash():
    hash_files = []
    for root, dirs, files in os.walk(brotools_path, topdown=False):
        for filename in files:
            name, ext = os.path.splitext(filename)
            if ext.lower() == ".py":
                filepath = os.path.join(root, filename)
                relpath = os.path.relpath(filepath, brotools_path)
                hash_files.append(filepath)
                # print (relpath, file_hash(filepath))

    hash_obj = hashlib.sha1(open(hash_files[0], 'rb').read())
    for fname in hash_files[1:]:
        hash_obj.update(open(fname, 'rb').read())
    checksum = hash_obj.hexdigest()
    return checksum


def individual_files_hash():
    hash_files = []
    hash_data = []
    for root, dirs, files in os.walk(brotools_path, topdown=False):
        for filename in files:
            name, ext = os.path.splitext(filename)
            if ext.lower() == ".py":
                filepath = os.path.join(root, filename)
                relpath = os.path.relpath(filepath, brotools_path)
                hash_files.append(filepath)
                hash_data.append([relpath, file_hash(filepath)])

    return hash_data


def file_hash(filepath):
    hash_obj = hashlib.sha1(open(filepath, 'rb').read())
    checksum = hash_obj.hexdigest()
    return checksum


# region Log debugging-related system information


import platform
import locale
from datetime import datetime

print("\n// -----------------")
print("{} Startup Log".format(MOD_NAME))
print("Loader Plugin Version: {}.{}\n".format(__version__, __release_date__))
print("-  {:<15} {}, {}".format("Log path:", str(log_path), str(cmdFileOutput_descriptor)))
print("-  {:<15} {}".format("Date started:", str(datetime.now())))
print("\nMAYA INFO")
print("-  {:<15} {}".format("OS:", cmds.about(q=True, os=True)))
print("-  {:<15} {}".format("UI Lang:", cmds.about(q=True, uil=True)))
print("-  {:<15} {}".format("Maya:", cmds.about(q=True, v=True)))
print("-  {:<15} {}".format("Maya full:", cmds.about(q=True, iv=True)))
print("-  {:<15} {}".format("Plugins:", cmds.pluginInfo(q=True, lsp=True)))
print("\nSYSTEM INFO")
print("-  {:<15} {}".format("Platform:", platform.platform()))
print("-  {:<15} {}".format("Locale Def:", locale.getdefaultlocale()))
print("-  {:<15} {}".format("Locale:", locale.getlocale()))
print("\nADDITIONAL INFO")
print("-  {:<15} {}".format("MOD_PATH:", MOD_PATH))
print("-  {:<15} {}".format("brotools_path:", brotools_path))
print("-  {:<15} {}".format("brotools_path files:", os.listdir(brotools_path)))
print("-  {:<15} {}".format("brotools hash:", brotools_hash()))
print("-  {:<15} {}".format("brotools files hash:", individual_files_hash()))

if cmds.about(q=True, win=True):
    import ctypes

    user32 = ctypes.windll.user32
    screensize = user32.GetSystemMetrics(0), user32.GetSystemMetrics(1), user32.GetSystemMetrics(78)

    print("-  {:<15} {}".format("Screen Info:", str(screensize)))
    print("-  {:<15} {}".format("Win:", str(sys.getwindowsversion())))

print("""\nIf you wish to keep any of this information private you're free to remove
some lines from here, however it can make debugging process harder. One part of
debugging process is to recreate your error. If I can't do that in my main system
my next step is to start a Virtual Machine with your OS and environment as close to
yours as possible. That's what all the above data is used for. \n""")
print("----------------- //\n")


# endregion

# region # # # # # # # # # # # # # # # # # # UTILITY FUNCTIONS # # # # # # # # # # # # # # # # # #


def clear_log():
    with codecs.open(os.path.join(brotools_write_path, 'bro_startup.log'), 'w', encoding='utf-8') as f:
        f.write('')


def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])

    return_parts = []
    for part in allparts:
        if part != os.path.sep and part not in [r"\\", "\\", "/", "//"]:
            return_parts.append(part)

    return return_parts


def log(*args, **kwargs):
    debug = kwargs.get('debug', False)
    if debug == True and VERBOSE_LOG == False:
        return

    tb = kwargs.get('traceback', False)
    critical = kwargs.get('critical', False)

    # time = str(datetime.now().strftime("%Y.%m.%d %H:%M:%S"))
    text = '// {}.py -'.format(MOD_NAME)
    for item in args:
        text += ' '
        text += str(item)
    text += " //"
    print(text)

    # with codecs.open(os.path.join(MOD_PATH, 'BroTools', 'bro_startup.log'), 'a+', encoding='utf-8') as f:
    #    f.write(text + "\n")

    if tb:
        cmds.warning(text)
        print("\n===== TRACEBACK START (SEND THIS TO DEVELOPER, you may also include log above) =====")
        traceback.print_exc()
        print("===== TRACEBACK END =====\n")

    if critical:
        cmds.error(text)


def catch_errors_critical(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except:
            log("Error in {} function!".format(f.__name__), traceback=True, critical=True)

    return wrapped


def catch_errors(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except:
            log("Error in {} function!".format(f.__name__), traceback=True)

    return wrapped


def rreload(module, depth=0, max_depth=3, top=True):
    """Recursively reload modules."""

    if top:
        log("rreload module", module, 'started...')

    if depth > max_depth:
        return

    for attribute_name in dir(module):
        if attribute_name != MOD_NAME and attribute_name != 'plug-ins':
            try:
                attribute = getattr(module, attribute_name)
                if type(attribute) is ModuleType:
                    print("=" * depth, attribute_name)
                    rreload(attribute, depth=depth + 1, top=False)
                    # reload(module)
                    mel.eval("reload({})".format(attribute_name))

            except:
                log("Unable to rreload", attribute_name)
    if top:
        log("rreload finished.")


# endregion


# Initialize MOD_NAME paths and setup
@catch_errors_critical
def setup_package_path(path):
    """
    Setup the packages that have been decoupled from the MOD.

    :param path: Path to add
    :return: Module parent directory.
    """
    for sysPath in sys.path:
        if path == sysPath:
            log("Removing", path, "from path.")
            sys.path.remove(path)

    if path not in sys.path:
        log("Appending", path, "to path.")
        sys.path.append(path)


def plugin_init_dummy():
    log("plugin_init_dummy")


# # Build menu
# @catch_errors
# def bro_menu_build(*args, **kwargs):
#     '''
#     This function builds the base menu for this module, and then walks through all subfolders of mod_dir, looking for
#       "menu.py" files, and runs their "build" function, allowing each subfolder to have it's own menu.
#
#     :param mod_dir: Folder to walk through and find menu.py files.
#     '''
#
#     log("{} Root Menu build...".format(MOD_Prefix))
#     try:
#         mod_dir = MOD_PATH
#
#         if mod_dir != None and mod_dir != "":
#
#             mod_dir = os.path.join(mod_dir, MOD_NAME)
#
#             # Check if menu exists and delete it
#             if cmds.menu("{}MenuRoot".format(MOD_Prefix), q=True, exists=True, aob=True):
#                 log("{}MenuRoot exists, remove.".format(MOD_Prefix))
#                 cmds.deleteUI("{}MenuRoot".format(MOD_Prefix))
#
#             # Create a new menu
#             if cmds.window('MayaWindow', exists=True):
#                 log("MayaWindow exists, create menu root.".format(MOD_Prefix))
#                 cmds.menu("{}MenuRoot".format(MOD_Prefix), l=MOD_NAME, p="MayaWindow", to=True)
#
#             # Build dynamic menus
#             log("Scan for menu files inside {}".format(mod_dir))
#             for root, dirs, files in os.walk(mod_dir, topdown=True):
#                 for name in files:
#                     fullpath = os.path.join(root, name)
#                     # log("   Scanning for menu:", str(fullpath), debug=True)
#                     if name.lower() == 'menu.py':
#                         try:
#                             log("   Building menu from file:", fullpath, debug=True)
#                             rel_path = os.path.relpath(fullpath, mod_dir)  # fullpath.replace(mod_dir, "")
#                             path_chunks = splitall(rel_path)
#
#                             modpath = MOD_NAME + "." + os.path.join(*path_chunks).replace(".py", "").replace(
#                                 os.path.sep,
#                                 ".")
#
#                             log("       Menu module, import and build:", modpath, debug=True)
#
#                             m = importlib.import_module(modpath)
#                             # reload(m)  # TODO Check if this should be returned
#                             m.build()
#                             log("      ", modpath, "menu built", debug=True)
#                         except:
#                             log("       x Unable to build menu from file:", modpath, traceback=True)
#
#             # Clear empty menus
#             from BroTools.configuration import optional_menu_ids
#             log("Cleaning up empty menus...", debug=True)
#             for menu_id in optional_menu_ids:
#
#                 try:
#                     menu_item_array = cmds.menu(menu_id, q=True, itemArray=True)
#                     do_cleanup = True if (menu_item_array == None or len(menu_item_array) < 1) else False
#                     log(menu_id, "clean up:", do_cleanup, debug=True)
#                     if do_cleanup:
#                         cmds.deleteUI(menu_id)
#                 except Exception as e:
#                     log("Can't clean up menu", menu_id, str(e))
#
#
#         else:
#             log("mod_dir not supplied for bro_menu_build. Skip.")
#             log(args, kwargs)
#     except:
#         log("Unknown error during menu build.", traceback=True)
#

# Remove menu
# @catch_errors_critical
# def bro_menu_build_remove():
#     '''
#     Removes menu
#     '''
#     log(MOD_NAME + " menu removing...")
#     if cmds.menu("{}MenuRoot".format(MOD_Prefix), q=True, exists=True, aob=True):
#         cmds.deleteUI("{}MenuRoot".format(MOD_Prefix))


def mel_import():
    """
    This is needed so that BroTools module is registered inside Maya for use with things like scriptJobs.
    """
    try:
        py_commands = [
            "import sys",
            "sys.path.append('{}')".format(MOD_PATH),
            "import {MOD_NAME}".format(MOD_NAME=MOD_NAME)
        ]
        for command in py_commands:
            mel.eval('''python (encodeString("{}"));'''.format(command))
    except:
        cmds.confirmDialog(title='BroTools: Startup error', message=r'''BroTools is unable to launch. Please make sure you followed all installation steps.

        To make sure everything works, please, follow these steps exactly:

        1. Close Maya
        2. Delete BroTools folder completely.
        3. Search your PC for 'BroTools', and if you find any folders or files - remove those.
        4. Find Maya.env file (Usually under C:\Users\USERNAME\Documents\maya\VERSION) anything related to BroTools from it
        4. Re-Download BroTools archive to make sure you got the latest one
        5. Unpack it in any folder on your PC. Make sure that path contains only latin letters and underscores. Using other symbols may interfere with a lot of programs' work. You can use your \Documents\maya folder if you wish so it will look like "\Documents\maya\BroTools"
        6. Open Maya
        7. Drag and drop "install.mel" into Maya viewport
        8. Restart Maya

        BroTools menu should appear at the top menu bar. You can also manually load\unload BroTools from Maya's plugin manager window.
        You may be able to find some information in Maya's Script Editor, it should display some error details there.

        If you still have troubles - contact me at support@brotools.tech and provide log files (brotools.log, bro_startup.log). You can find them under BroTools folder.

        Cheers!''', icon="critical")
        log(r"{} mel eval import failed!".format(MOD_NAME), traceback=True, critical=True)


def on_register_ui(*args, **kwargs):
    from BroTools.core.system import utilities as su
    # bro_menu_build()
    # mel_import()

    log(MOD_NAME + " init finished...")
    log("Close startup log file...")
    close_result = cmds.cmdFileOutput(c=cmdFileOutput_descriptor)
    if close_result != 0:
        log(r"Unable to close startup log file: ", close_result)
        log(
            r"You can read about error codes here: http://download.autodesk.com/us/maya/2011help/CommandsPython/cmdFileOutput.html")

    try:
        su.on_plugin_initialized()
    except Exception as e:
        cmds.warning("on_plugin_initialized failed: {}".format(e))
        traceback.print_exc()


def on_unregister_ui(*args, **kwargs):
    from BroTools.core.system import utilities as su
    su.on_plugin_uninitialized()
    # bro_menu_build_remove()


# Initialize the script plug-in
@catch_errors_critical
def initializePlugin(mobject):
    '''
    Used by Maya to initialize Maya plugin
    :param mobject: Passed by Maya.
    :return: Returns status of registerUI function.
    '''
    # clear_log()
    log("{} Plugin initializing...".format(MOD_NAME))

    # Remove pyc files on startup to make sure we're using latest versions after updates, and to avoid the possible
    # issue with using 2 maya versions with python 2 and 3.
    remove_pyc_files()

    mplugin = OpenMayaMPx.MFnPlugin(mobject, "Michael Davydov\nhttp://michaeldavydov.com", "1.0")

    # Setup python, script and plugin paths
    setup_package_path(MOD_PATH)
    for pt in PackagePaths:
        setup_package_path(os.path.join(MOD_PATH, "BroTools", *pt))

    # This gets registered here, but only executed AFTER initializePlugin.
    status = mplugin.registerUI(on_register_ui, on_unregister_ui)
    log("initializePlugin end with status", status)

    return status


# Uninitialize the script plug-in
def uninitializePlugin(mobject):
    '''
    Used by Maya to uninitialize plugin.
    :param mobject: Passed by Maya.
    '''
    log(MOD_NAME + " Plugin unloading...")
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    # plugin_path = cmds.pluginInfo(MOD_NAME + '.py', query=True, path=True)
    # plugin_dir = os.path.dirname(plugin_path)
    #
    # MOD_dir = os.path.join(plugin_dir, '..', '..')

    log("Cleaning up windows...")
    import BroTools.core.broqt.bro_windows as bw
    for w in bw.instance_cache:
        try:
            w.setParent(None)
            w.deleteLater()
        except:
            pass

    bw.instance_cache = []

    log("Unloading modules...")
    userPath = MOD_PATH.lower()
    modules = []
    for root, dirs, files in os.walk(MOD_PATH, topdown=False):
        for name in files:
            if name in ['.py', '.pyd', '.so'] and "__init__" not in name and ".pyc" not in name:
                modulePath = os.path.join(root, name)
                modulePath = modulePath.replace(MOD_PATH + os.path.sep, "")
                modulePath = modulePath.replace(os.path.sep, ".")
                modulePath = MOD_NAME + "." + modulePath.replace('.py', '')
                modules.append(modulePath)

    for key, module in sys.modules.items():
        # There's a few modules that are going to complain if you try to query them
        # so I've popped this into a try/except to keep it safe
        try:
            # Use the "inspect" library to get the moduleFilePath that the current module was loaded from
            moduleFilePath = inspect.getfile(module).lower()

            # Don't try and remove the startup script, that will break everything
            if moduleFilePath == plugin_path.lower():
                continue

            # If the module's filepath contains the userPath, add it to the list of modules to delete
            if moduleFilePath.startswith(userPath):
                modules.append(key)
        except:
            pass

    print("BroTools modules:\n    - ", "\n     - ".join(modules))
    success = []
    error = []
    for module in reversed(modules):
        try:
            # del sys.modules[module]
            # reload(sys.modules[module])
            # log("REloaded", module)
            del (sys.modules[module])
            if module not in success:
                success.append(module)
                # log("UNloaded", module)
        except Exception as e:
            if module not in error:
                error.append(str(module) + ": " + str(e))
                # log("SKIP", module)

                # for er in error:
                # log("Skipped", module)
                # log(er)
                # print "\n\n"
    print("Unloaded modules:\n    - ", "\n     - ".join(success))
    print("Failed unloading:\n    - ", "\n     - ".join(error))
