from base64 import b64decode as license64

import unittest, os
from maya import cmds
from BroTools.configuration import brotools_config
FILE_DIR = os.path.dirname(os.path.realpath(__file__))
TESTS_DIR = FILE_DIR
TEMP_DIR = os.path.join(FILE_DIR)
KEEP_FILES = False
BUFFER_OUTPUT = True
FILE_NEW = False

class TestCase(unittest.TestCase):
    'Base class for unit test cases run in Maya.\n\n    Tests do not have to inherit from this TestCase but this derived TestCase contains convenience\n    functions to load/unload plug-ins and clean up temporary files.\n    '
    files_created = []
    plugins_loaded = set()

    @classmethod
    def tearDownClass(A):
        super(TestCase, A).tearDownClass()
        A.delete_temp_files()
        A.unload_plugins()

    @classmethod
    def load_plugin(B, plugin):
        'Load the given plug-in and saves it to be unloaded when the TestCase is finished.\n\n        @param plugin: Plug-in name.\n        '
        A = plugin
        cmds.loadPlugin(A, qt=True)
        B.plugins_loaded.add(A)

    @classmethod
    def unload_plugins(A):
        for B in A.plugins_loaded:
            cmds.unloadPlugin(B)
        A.plugins_loaded = []

    @classmethod
    def delete_temp_files(A):
        'Delete the temp files in the cache and clear the cache.'
        if (not KEEP_FILES):
            for B in A.files_created:
                if os.path.exists(B):
                    os.remove(B)
            A.files_create = []

    @classmethod
    def get_temp_filename(F, file_name):
        "Get a unique filepath name in the testing directory.\n\n        The file will not be created, that is up to the caller.  This file will be deleted when\n        the tests are finished.\n        @param file_name: A partial path ex: 'directory/somefile.txt'\n        @return The full path to the temporary file.\n        "
        A = TEMP_DIR
        if (not os.path.exists(A)):
            os.makedirs(A)
        (C, D) = os.path.splitext(file_name)
        B = '{0}/{1}{2}'.format(A, C, D)
        E = 0
        while os.path.exists(B):
            E += 1
            B = '{0}/{1}{2}{3}'.format(A, C, E, D)
        F.files_created.append(B)
        return B

    @classmethod
    def open_scene(B, filepath=None):
        A = filepath
        if (A is not None):
            cmds.file(A, f=True, ignoreVersion=True, o=True)
        else:
            cmds.file(f=True, new=True)

class DebugLoggingState(object):
    initial_state = None

    @classmethod
    def suppress_output(A):
        if BUFFER_OUTPUT:
            A.initial_state = brotools_config.Debug.debug_level
            brotools_config.Debug.debug_level = 10
            brotools_config.save()

    @classmethod
    def restore_output(A):
        if (A.initial_state is not None):
            brotools_config.Debug.debug_level = A.initial_state
            brotools_config.save()

class ScriptEditorState(object):
    'Provides methods to suppress and restore script editor output.'
    suppress_results = None
    suppress_errors = None
    suppress_warnings = None
    suppress_info = None

    @classmethod
    def suppress_output(A):
        'Hides all script editor output.'
        if BUFFER_OUTPUT:
            A.suppress_results = cmds.scriptEditorInfo(q=True, suppressResults=True)
            A.suppress_errors = cmds.scriptEditorInfo(q=True, suppressErrors=True)
            A.suppress_warnings = cmds.scriptEditorInfo(q=True, suppressWarnings=True)
            A.suppress_info = cmds.scriptEditorInfo(q=True, suppressInfo=True)
            cmds.scriptEditorInfo(e=True, suppressResults=True, suppressInfo=True, suppressWarnings=True, suppressErrors=True)

    @classmethod
    def restore_output(A):
        'Restores the script editor output settings to their original values.'
        if (None not in {A.suppress_results, A.suppress_errors, A.suppress_warnings, A.suppress_info}):
            cmds.scriptEditorInfo(e=True, suppressResults=A.suppress_results, suppressInfo=A.suppress_info, suppressWarnings=A.suppress_warnings, suppressErrors=A.suppress_errors)

class TestResult(unittest.TextTestResult):
    'Customize the test result so we can do things like do a file new between each test and suppress script\n    editor output.\n    '

    def startTestRun(A):
        'Called before any tests are run.'
        super(TestResult, A).startTestRun()
        ScriptEditorState.suppress_output()
        DebugLoggingState.suppress_output()

    def stopTestRun(A):
        'Called after all tests are run.'
        ScriptEditorState.restore_output()
        DebugLoggingState.restore_output()
        super(TestResult, A).stopTestRun()

    def stopTest(A, test):
        'Called after an individual test is run.\n\n        @param test: TestCase that just ran.'
        super(TestResult, A).stopTest(test)
        if FILE_NEW:
            cmds.file(f=True, new=True)

def run():
    A = unittest.TestLoader()
    B = TESTS_DIR
    C = A.discover(B)
    D = unittest.TextTestRunner(verbosity=2, resultclass=TestResult)
    D.run(C)
